/*****************************************************************************
#   Copyright (C) 1994-1999 by Phil Green.                          
#   All rights reserved.                           
#                                                                           
#   This software is part of a beta-test version of the swat/cross_match/phrap 
#   package.  It should not be redistributed or
#   used for any commercial purpose, including commercially funded
#   sequencing, without written permission from the author and the
#   University of Washington.
#   
#   This software is provided ``AS IS'' and any express or implied
#   warranties, including, but not limited to, the implied warranties of
#   merchantability and fitness for a particular purpose, are disclaimed.
#   In no event shall the authors or the University of Washington be
#   liable for any direct, indirect, incidental, special, exemplary, or
#   consequential damages (including, but not limited to, procurement of
#   substitute goods or services; loss of use, data, or profits; or
#   business interruption) however caused and on any theory of liability,
#   whether in contract, strict liability, or tort (including negligence
#   or otherwise) arising in any way out of the use of this software, even
#   if advised of the possibility of such damage.
#                                                                         
#*****************************************************************************/

#include "swat.h"
#include "snp.h"
#include <string.h>

#define MAX_QUAL 99

#define EXTENT_FUDGE 7 /* adjustment to extent segments to allow for spurious extensions
			  to extents -- also in pairs.c */

#define MIN_QUAL_SEG_SIZE 30 /* minimum size of high-quality region -- if less than this, then
				is reset to size 0 */

extern Parameters *parameters;
extern int t_num_entries;

static double **LLR_discrep[2], **LLR_match[2];
static int **disc_qual_matrix[2], **match_qual_matrix[2]; /* used for computing log_likelihoods of
						     discrepancies, cond'l on qualities */
static int *qual_LLR;
static double *err_rate, *qual_freq;
static double ref_conserve, match_credit;
static int mismatch_LLR[256];
static int scaled_err_probs[256]; /* used for finding high-quality segments */

init_qual_arrays()
{
  int i;
  double x, log_ref, log95;
  char *our_alloc();

  ref_conserve = parameters->repeat_stringency;
  log_ref = log10(1. / ref_conserve);
  log95 = log10(1. / .95); /* this is scaling factor -- so that negative cutoffs retain
			      similar meaning */
  match_credit = 10 * log95; /* log_ref; */
  x = -log(10.0) / 10.0;
/*
  fprintf(stderr, "\nDiscrepancy penalties, by min quality -- repeat_stringency = %.3f: ",
	  ref_conserve);
*/
  for (i = 0; i < 99; i++) {
    mismatch_LLR[i] = (log95 / log_ref) * (-10 * log10(1.0 - ref_conserve + ref_conserve * exp(i * x)) - i);
/*
    fprintf(stderr, "\n%2d  %d ", i, mismatch_LLR[i]);
*/
  }
  for (; i < 256; i++) 
    mismatch_LLR[i] = FORCE_REJECT_SCORE; /* to force marked discrepant reads (having qualities of 99) 
					     not to overlap */

/* following is temporarily inactivated 
  qual_LLR  = (int *)our_alloc((MAX_QUAL + 1) * sizeof(int));
  err_rate  = (double *)our_alloc((MAX_QUAL + 1) * sizeof(double));
  qual_freq  = (double *)our_alloc((MAX_QUAL + 1) * sizeof(double));
  for (k = 0; k < 2; k++) {
    disc_qual_matrix[k] = (int **)our_alloc((MAX_QUAL + 1) * sizeof(int *));
    match_qual_matrix[k] = (int **)our_alloc((MAX_QUAL + 1) * sizeof(int *));
    LLR_discrep[k] = (double **)our_alloc((MAX_QUAL + 1) * sizeof(double *));
    LLR_match[k] = (double **)our_alloc((MAX_QUAL + 1) * sizeof(double *));
    for (i = 0; i <= MAX_QUAL; i++) {
      disc_qual_matrix[k][i] = (int *)our_alloc((MAX_QUAL + 1) * sizeof(int));
      match_qual_matrix[k][i] = (int *)our_alloc((MAX_QUAL + 1) * sizeof(int));
      LLR_discrep[k][i] = (double *)our_alloc((MAX_QUAL + 1) * sizeof(double));
      LLR_match[k][i] = (double *)our_alloc((MAX_QUAL + 1) * sizeof(double));
    }
  }    
*/
} 

set_qual_arrays(db)
     Database *db;
{
  int i, j, k, entry1, reverse, t_n_disc, t_n_match, n_disc, n_match;
  int site1, site2, q1, q2, last_disc1, last_disc2, length2;
  char *diff, *qual1, *qual2;
  double quot, eij;
  Aligned_pair *pair;
  Aligned_pair *get_aligned_pairs();
  int t_pairs[2];
  char *get_adj_qual();

/* check that qualities are in bounds -- this should be a separate routine */
  for (entry1 = db->first_entry; entry1 <= db->last_entry; entry1++) {
    qual1 = get_adj_qual(entry1);
    for (i = 0; i < get_seq_length(entry1); i++)
      if (qual1[i] > MAX_QUAL || qual1[i] < 0) {
	fprintf(stderr, "\nQuality out of bounds: %d, max_qual: %d ", 
		qual1[i], MAX_QUAL);
	exit(1);
      }
  }

 /* reference repeat */
  for (i = 0; i <= MAX_QUAL; i++) 
    qual_LLR[i] = qual_freq[i] = err_rate[i] = 0;

  for (k = 0; k < 2; k++) 
    for (i = 0; i <= MAX_QUAL; i++) 
      for (j = 0; j <= MAX_QUAL; j++)
	 disc_qual_matrix[k][i][j] = match_qual_matrix[k][i][j] = 
	   LLR_discrep[k][i][j] = LLR_match[k][i][j] = 0;

  for (entry1 = db->first_entry; entry1 <=  db->last_entry; entry1++) {
/*    if (is_anomalous(align_entry1)) continue; */
    qual1 = get_adj_qual(entry1);
    for (pair = get_aligned_pairs(entry1); pair; pair = pair->next) {
      if (pair->entry1 >= pair->entry2) continue;
      if (pair->LLR_score < 0) continue;
/*
      if (pair->score < parameters->confirm_score) continue;
      if (!is_best(pair)) continue;
      if (is_reject_any(pair)) continue;
*/
/*      if (is_anomalous(align_entry2)) continue; */
      qual2 = get_adj_qual(pair->entry2);
      reverse = is_reverse(pair);
      length2 = get_seq_length(pair->entry2) - 1;
      last_disc1 = site1 = pair->start1 - 1;
      last_disc2 = site2 = pair->start2 - 1;
      for (diff = pair->diffs; *diff; diff++) {
	site1 += diff_gap1(*diff);
	site2 += diff_gap2(*diff);
	if (diff_type(*diff) == 'M') continue;
	for (i = last_disc1 + 1, j = last_disc2 + 1; i < site1 || j < site2; i++, j++) {
	  q1 = qual1[i];
	  q2 = qual2[reverse ? length2 - j : j];
	  match_qual_matrix[reverse][q1][q2] += 1;
	  match_qual_matrix[reverse][q2][q1] += 1;
	}
	if (!*(diff + 1)) break;
	last_disc1 = site1;
	last_disc2 = site2;
	q1 = qual1[site1];
	q2 = qual2[reverse ? length2 - site2 : site2];
	disc_qual_matrix[reverse][q1][q2] += 1;
	disc_qual_matrix[reverse][q2][q1] += 1;
      }
    }
  }

  t_n_disc = t_n_match = 0;
  t_pairs[0] = t_pairs[1] = 0;

  for (i = 1; i <= MAX_QUAL; i++) {
    n_disc = n_match = 0;
    for (j = 1; j <= MAX_QUAL; j++) {
      n_disc += disc_qual_matrix[0][i][j] + disc_qual_matrix[1][i][j];
      n_match += match_qual_matrix[0][i][j] + match_qual_matrix[1][i][j];
      t_pairs[0] += disc_qual_matrix[0][i][j] + match_qual_matrix[0][i][j];
      t_pairs[1] += disc_qual_matrix[1][i][j] + match_qual_matrix[1][i][j];
    }
    t_n_disc += n_disc;
    t_n_match += n_match;
    qual_freq[i] = n_disc + n_match;
    err_rate[i] = (n_disc + 1.0) / (n_disc + n_match + 1.0);
  } 

  quot = .5 * (t_n_disc + 1.0) / (t_n_disc + t_n_match + 1.0);
  printf("\n\nQual freq (%%), err_rate (%%), LLR:");
  for (i = 1; i <= MAX_QUAL; i++) {
    if (!qual_freq[i]) continue;
    qual_freq[i] /= t_n_disc + t_n_match;
    err_rate[i] -= quot;
    if (err_rate[i] < .0001) err_rate[i] = .0001;
    qual_LLR[i] = 10 * log10(err_rate[i]);
    printf("\n%2d   %6.2f  %6.2f  %6.1f", i, 100.0 * qual_freq[i], 
	   100.0 * err_rate[i], qual_LLR[i] / 10.0);
  }

  for (k = 0; k < 2; k++)  
    for (i = 1; i <= MAX_QUAL; i++) 
      for (j = 1; j <= MAX_QUAL; j++) {
/* version 950810
	quot = (1.0 + disc_qual_matrix[k][i][j]) / 
	  (1.0 + disc_qual_matrix[k][i][j] + match_qual_matrix[k][i][j]);
	LLR_discrep[k][i][j] = 10 * log10(quot / (ref_conserve * quot + 1 - ref_conserve));
*/
	eij = ref_conserve * (1.0 - err_rate[i]) * (1.0 - err_rate[j]);
	if (disc_qual_matrix[k][i][j] + match_qual_matrix[k][i][j]) {
	  LLR_discrep[k][i][j] = 
	    10 * log10(((1.0 + disc_qual_matrix[k][i][j]) / (1.0 + t_pairs[k]))
		       / (qual_freq[i] * qual_freq[j] * (1 - eij)));
	  if (LLR_discrep[k][i][j] > 0) LLR_discrep[k][i][j] = 0; 
/*
	  LLR_match[k][i][j] = 
	    10 * log10(((1.0 + match_qual_matrix[k][i][j]) / (1.0 + t_pairs[k]))
		       / (qual_freq[i] * qual_freq[j] * eij));
	  if (LLR_match[k][i][j] < 0) LLR_match[k][i][j] = 0; 
*/
	  LLR_match[k][i][j] = match_credit;
	}
      }
 
  printf("\n\nRead/read discrepancies, by quality:");
  printf("\n  ");
  for (i = 0; i <= MAX_QUAL; i++) 
    if (qual_freq[i]) printf("  %5d", i);
  
  for (k = 0; k < 2; k++) {
    printf("\n          %s strand", k ? "Opposite" : "Same");
    for (i = 0; i <= MAX_QUAL; i++) {
      if (qual_freq[i]) {
	printf("\n%2d", i);
	for (j = 0; j <= i; j++) 
	  if (qual_freq[j]) printf("  %5d", disc_qual_matrix[k][i][j]);
      }
    }
  }

  printf("\n\nPer cent:");
  for (k = 0; k < 2; k++) {
    printf("\n          %s strand", k ? "Opposite" : "Same");
    for (i = 0; i <= MAX_QUAL; i++) {
      if (!qual_freq[i]) continue;
      printf("\n%2d", i);
      for (j = 0; j <= i; j++) {
	if (!qual_freq[j]) continue;
	quot = disc_qual_matrix[k][i][j] + match_qual_matrix[k][i][j];
	if (!quot) quot = 1.0;
	printf(" %6.3f", (100.0 * disc_qual_matrix[k][i][j]) / quot);
      }
    }
  }

  printf("\n\nLLR_discrep matrix (for %.2f repeat):", ref_conserve);
  for (k = 0; k < 2; k++) {
    printf("\n          %s strand", k ? "Opposite" : "Same");
    for (i = 0; i <= MAX_QUAL; i++) {
      if (!qual_freq[i]) continue;
      printf("\n%2d", i);
      for (j = 0; j <= i; j++) 
	if (qual_freq[j]) printf(" %6.3f", LLR_discrep[k][i][j] / 10.0);
    }
  }

  printf("\n\nLLR_match matrix (for %.2f repeat):", ref_conserve);
  for (k = 0; k < 2; k++) {
    printf("\n          %s strand", k ? "Opposite" : "Same");
    for (i = 0; i <= MAX_QUAL; i++) {
      if (!qual_freq[i])  continue;
      printf("\n%2d", i);
      for (j = 0; j <= i; j++) 
	if (qual_freq[j]) printf(" %6.3f", LLR_match[k][i][j] / 10.0);
    }
  }
}

/* compute LLR scores; BUT if first_file_only, just make them equal to raw swat score
   (this should be changed, at some point) */

set_LLR_scores()
{
  Aligned_pair *pair;
  Aligned_pair *get_aligned_pairs();
  int entry1, i, cum, score;
  int LLR_hist[400];
  char *o_qual1, *a_qual1;
  char *get_adj_qual(), *get_orig_qual();
  Align_info *get_align_entry();
  Segment *segments1, *segments2;
  char *get_id();

  notify("Computing LLR scores ...");
  for (i = 0; i < 400; i++) LLR_hist[i] = 0;

  for (entry1 = 0; entry1 < t_num_entries; entry1++) {
    o_qual1 = get_orig_qual(entry1);
    a_qual1 = get_adj_qual(entry1);
    segments1 = get_align_entry(entry1)->segments;
    for (pair = get_aligned_pairs(entry1); pair; pair = pair->next) {
      if (entry1 >= pair->entry2) continue;
      segments2 = get_align_entry(pair->entry2)->segments;
      score = pair->LLR_score = pair->reversed_pair->LLR_score = 
	parameters->subject_files ? pair->score :
	  get_LLR(pair->diffs, o_qual1, get_orig_qual(pair->entry2), 
		  a_qual1, get_adj_qual(pair->entry2), 
		  get_seq_length(entry1), pair->start1, pair->end1, 
		  get_seq_length(pair->entry2), pair->start2, pair->end2, 
		  is_reverse(pair), segments1, segments2, 0, 0, 0, 0, 0, 0, 0);
      score = (score / 50) + 300;
      score = score < 0 ? 0 : (score > 399 ? 399 : score);
      LLR_hist[score] += 1;
    }
  }
  printf("\n\nLLR score histogram:\nScore    #   cum # ");
  for (i = cum = 0; i < 400; i++) {
    if (LLR_hist[i]) {
      cum += LLR_hist[i];
      printf("\n%5.1f  %4d  %4d", (i - 300) * 5.0, LLR_hist[i], cum);
    }
  }
  notify(" Done\n");
} 

get_site_LLR(q_val)
     int q_val;
{
  return qual_LLR[q_val];
}

#define MAX_UNALIGNED 49
#define CELL_SIZE 3

get_LLR(diffs, orig_qual1, orig_qual2, adj_qual1, adj_qual2, length1, start1, end1, length2, start2, end2, reverse, segments1, segments2, print_flag, fp, q_start1, q_end1, q_start2, q_end2, ignore_ends)
     char *diffs, *orig_qual1, *orig_qual2, *adj_qual1, *adj_qual2;
     int length1, start1, end1, length2, start2, end2, reverse;
     Segment *segments1, *segments2;
     int print_flag;
     FILE *fp;
     int q_start1, q_end1, q_start2, q_end2;
     int ignore_ends;
{
  int site1, site2, q1, q2, last_disc1, last_disc2;
  int i, j, unaligned;
  int n_matches;
  double score, score_lt_20, score_gt_20, score_in_HQ, score_out_HQ;
  int score_vector[MAX_UNALIGNED];
  int get_unaligned_LLR();
  char *diff;
  int d_score;
  Segment *segment;
  int num_lt_20, num_gt_20;

/* following are necessary, since init_qual_arrays no longer used */

  n_matches = score = score_lt_20 = score_gt_20 = score_in_HQ = score_out_HQ = 0;
  num_lt_20 = num_gt_20 = 0;
  length1--;
  length2--;
  last_disc1 = site1 = start1 - 1;
  last_disc2 = site2 = start2 - 1;
  for (diff = diffs; *diff; diff++) {
    site1 += diff_gap1(*diff);
    site2 += diff_gap2(*diff);
    if (diff_type(*diff) == 'M') continue;
    for (i = last_disc1 + 1, j = last_disc2 + 1; i < site1 || j < site2; i++, j++) {
/*
      if ((q1 = adj_qual1[i]) && (q2 = adj_qual2[reverse ? length2 - j : j])) 
	score += LLR_match[reverse][q1][q2]; 
      if ((q1 = orig_qual1[i]) && (q2 = orig_qual2[reverse ? length2 - j : j])) 
	score += match_credit; 
*/
      if (adj_qual1[i] && adj_qual2[reverse ? length2 - j : j]) 
	n_matches++;

    }
						
/* 950810	n_matches++; */
    if (!*(diff + 1)) break;
    last_disc1 = site1;
    last_disc2 = site2;
    q1 = adj_qual1[site1];
    q2 = adj_qual2[reverse ? length2 - site2 : site2];
    if ((d_score = discrep_score(q1, q2)) == FORCE_REJECT_SCORE)
      return FORCE_REJECT_SCORE;

    score += d_score; 
    if (print_flag) {
      if (d_score > -20) {
	score_lt_20 += d_score;
	num_lt_20++;
      }
      else {
	score_gt_20 += d_score;
	num_gt_20++;
      }
      if (site1 >= q_start1 && site1 <= q_end1 && 
	  site2 >= q_start2 && site2 <= q_end2) 
	score_in_HQ += d_score;
      else score_out_HQ += d_score;
    }
/* this is approximation  */
/*
    score += LLR_discrep[reverse][q1][q2];
    q1 = orig_qual1[site1];
    q2 = orig_qual2[reverse ? length2 - site2 : site2];
    score -= q1 < q2 ? q1 : q2; 
*/

  }

  if (print_flag) {
    fprintf(fp,"\nLLR breakdown: discreps: %.1f (<20 part: %.1f (#=%d), >20:%.1f (#=%d); in HQ: %.1f, out HQ %.1f), match: %.1f ", 
	    score/10.0, score_lt_20/10.0, num_lt_20, score_gt_20/10.0, num_gt_20, score_in_HQ/10.0, score_out_HQ/10.0, (double)n_matches * match_credit/10.0);
  }
  score += n_matches * match_credit;

/* 950810	  score += n_matches * 10 * log_ref; */
/* find minimum possible penalty for unaligned parts of reads */
  unaligned = MAX_UNALIGNED; /* look at maximum of 33 bases at each end (8 "cells") */
  if (unaligned > length1 - end1) unaligned = length1 - end1;
/* adjust for chimeric reads */
  if (segments1 && segments1->next) {
    for (segment = segments1->next; segment; segment = segment->next) {
      if (segment->start + EXTENT_FUDGE > end1) {
	if (unaligned > segment->start - end1) 
	  unaligned = segment->start - end1;
      }
    }
  }
  if (unaligned > length2 - end2) unaligned = length2 - end2;
  if (segments2 && segments2->next) {
    if (reverse) {
      for (segment = segments2; segment->next; segment = segment->next) {
	if ((length2 - segment->end) + EXTENT_FUDGE > end2) {
	  if (unaligned > (length2 - segment->end) - end2) 
	    unaligned = (length2 - segment->end) - end2;
	}
      }
    }
    else {
      for (segment = segments2->next; segment; segment = segment->next) {
	if (segment->start + EXTENT_FUDGE > end2) {
	  if (unaligned > segment->start - end2) 
	    unaligned = segment->start - end2;
	}
      }
    }
  }

  if (unaligned > 0) {
    for (i = 0; i < unaligned; i++) {
      q1 = adj_qual1[i + end1 + 1];
      site2 = i + end2 + 1;
      q2 = adj_qual2[reverse ? length2 - site2 : site2];
      score_vector[i] = discrep_score(q1, q2);
/*
      score_vector[i] = LLR_discrep[reverse][q1][q2];
      q1 = orig_qual1[i + end1 + 1];
      q2 = orig_qual2[reverse ? length2 - site2 : site2];
      score_vector[i] -= q1 < q2 ? q1 : q2; 
*/
    }
    if ((d_score = get_unaligned_LLR(score_vector, unaligned)) == FORCE_REJECT_SCORE)
      return FORCE_REJECT_SCORE;
    if (print_flag) 
      fprintf(fp," trail: %.1f ", d_score/10.0);
    if (!ignore_ends) score += d_score;
  }

  unaligned = MAX_UNALIGNED;

  if (unaligned > start1) unaligned = start1;
  if (segments1 && segments1->next) {
    for (segment = segments1; segment->next; segment = segment->next) {
      if (segment->end - EXTENT_FUDGE < start1) {
	if (unaligned > start1 - segment->end) 
	  unaligned = start1 - segment->end;
      }
    }
  }
  if (unaligned > start2) unaligned = start2;
  if (segments2 && segments2->next) {
    if (reverse) {
      for (segment = segments2->next; segment; segment = segment->next) {
	if ((length2 - segment->start) - EXTENT_FUDGE < start2) {
	  if (unaligned > start2 - (length2 - segment->start)) 
	    unaligned = start2 - (length2 - segment->start);
	}
      }
    }
    else {
      for (segment = segments2; segment->next; segment = segment->next) {
	if (segment->end - EXTENT_FUDGE < start2) {
	  if (unaligned > start2 - segment->end) 
	    unaligned = start2 - segment->end;
	}
      }
    }
  }



  if (unaligned > 0) {
    for (i = 0; i < unaligned; i++) {
      q1 = adj_qual1[start1 - 1 - i];
      site2 = start2 - 1 - i;
      q2 = adj_qual2[reverse ? length2 - site2 : site2];
      score_vector[i] = discrep_score(q1, q2);
/*
      score_vector[i] = LLR_discrep[reverse][q1][q2];
      q1 = orig_qual1[start1 - 1 - i];
      q2 = orig_qual2[reverse ? length2 - site2 : site2];
      score_vector[i] -= q1 < q2 ? q1 : q2; 
*/
    }
    if ((d_score = get_unaligned_LLR(score_vector, unaligned)) == FORCE_REJECT_SCORE)
      return FORCE_REJECT_SCORE;
    if (!ignore_ends) score += d_score;
    if (print_flag) 
      fprintf(fp," lead: %.1f ", d_score/10.0);
  }
  if (print_flag) 
    fprintf(fp," total: %.1f ", score/10.0);
  return (int)score < FORCE_REJECT_SCORE + 1 ? FORCE_REJECT_SCORE + 1 : (int)score;
}

int discrep_score(q1, q2)
     int q1, q2;
{
  q1 = q1 < q2 ? q1 : q2; /* this is approximate! Should actually be slightly smaller */

  return mismatch_LLR[q1];
}


/*  get any negative score in unaligned region (keeping in mind that there must
    be enough mismatches there to make score non-positive; and assuming those
    mismatches are in most favorable positions with respect to LLR score */
get_unaligned_LLR(score_vector, unaligned)
     int *score_vector;
     int unaligned;
{
  int n_cells, i_cell, i;
  int score, temp;

  score = score_vector[0];
  n_cells = (unaligned - 1) / CELL_SIZE; /* here cell size = mismatch penalty + 1 (now avg. of indel and 
				    mismatch penalty is no. of positions
				    required to include one mismatch + compensating matches, 
				    assuming a match reward of 1. */
  for (i_cell = 1; i_cell <= n_cells; i_cell++) {
    /* since maximum score is 0, if attain it don't need to go any further */
/* do not allow marked regions to be displaced */
/* following removed 8/5/98 -- causes incorrect rejection of some merges.
   should be unnec with new consed.
    if (score_vector[i_cell] == FORCE_REJECT_SCORE) 
      return FORCE_REJECT_SCORE;
*/
    for (i = i_cell + 1; score_vector[i_cell] && i <= i_cell * CELL_SIZE; i++) 
      if (score_vector[i_cell] < (temp = score_vector[i])) { 
	score_vector[i] = score_vector[i_cell];
	score_vector[i_cell] = temp;
      }

    if (score_vector[i_cell] == FORCE_REJECT_SCORE) 
      return FORCE_REJECT_SCORE;
    score += score_vector[i_cell];
  }
  return score < FORCE_REJECT_SCORE + 1 ? FORCE_REJECT_SCORE + 1 : score;
}

/* read in quality data */
read_qual(db)
     Database *db;
{
  FILE *fp;
  FILE *fopen();
  File *file;
  int i, j, k, length, entry1, c;
  char *our_alloc();
  char *qf_name, *seq, *orig_qual, *id, *idq;
  char *get_seq(), *get_orig_qual();
  char *get_id();
  Align_info *align_entry;
  Align_info *get_align_entry();
/*
  char dig_buffer[1000];
*/
  int i_string;
  double trim_err, x;

  notify("Reading quality files ...");

  trim_err = pow(10.0, -parameters->trim_qual / 10.0);
  for (i = 0; i < 256; i++) {
    x = i < 100 ? pow(10.0, -i / 10.0) : 0.0;
    scaled_err_probs[i] = 1000.0 * (trim_err - x);
  }
  
  entry1 = db->first_entry;

  file = db->file; 
  qf_name = (char *)our_alloc((strlen(file->name) + 6) * sizeof(char));
  strcpy(qf_name, file->name);
  strcat(qf_name, ".qual");
  fp = fopen(qf_name, "r");
  init_qual_hist();
  if (fp) {
    parameters->qual_flag = 1;
    printf("\n\nQuality file: %s", qf_name);
    fgetc(fp);
    do {
      id = get_id(entry1);
      for (c = fgetc(fp), idq = id; c == *idq; 
	   c = fgetc(fp), idq++);
      if (!isspace(c) && c != '.') { /* 2d condition is to allow for trimming of .seq */
	fprintf(stderr,"\nFile %s, entry %s", qf_name, id);
	fatalError("Inconsistency between quality file and sequence file");
      }
      for (; c != '\n' && c != EOF; c = fgetc(fp));
      
      seq = get_seq(entry1);
      length = get_seq_length(entry1); 
      
      orig_qual = get_orig_qual(entry1);
      align_entry = get_align_entry(entry1);

      for (j = 0; j < length; j++) {
	for (; isspace(c); c = fgetc(fp));
	for (i_string = k = 0; isdigit(c); c = fgetc(fp), i_string++) 
	  k = 10 * k + (c - '0'); /* assumes ASCII convention */
	
	if (!i_string || k > MAX_QUAL) {
	  if (!i_string && !isspace(c))
	    fprintf(stderr, 
		    "Error: ASCII character %d found where digit expected in %s, entry %s: position %d out of %d", 
		    c, qf_name, get_id(entry1), j + 1, length);
	  else 
	    fprintf(stderr, 
		  "Error or quality value out of bounds (< 0 or > %d) in file %s, entry %s", 
		  MAX_QUAL, qf_name, get_id(entry1));
	  fatalError("");
	}
	orig_qual[j] = k != 98 ? k : 0; /* change to 0 for 'N' ?? */
      }
      
      while (isspace(c = fgetc(fp)));
      if (c != '>' && c != EOF) {
	fprintf(stderr,"\nFile %s, entry %s\n%c\n", qf_name, id, c);
	fatalError("Error or file inconsistency: quality and sequence file");
      }
      incr_qual_hist(orig_qual, length);

/* note if qualities are rescaled, this may not be what is desired */

      best_qual_seg(orig_qual, length, &(align_entry->qual_start), 
		    &(align_entry->qual_end));
      entry1++;
    } while (c != EOF); 
    fclose(fp);
    print_qual_hist("Input quality");
    our_free(qf_name);
  }
  else { 
    parameters->qual_flag = 0;
    printf("\n\nNO QUALITY FILE %s WAS FOUND. REMAINING INPUT QUALITIES SET TO %d.", 
	   qf_name, parameters->default_qual);
    fprintf(stderr, "\n\nNO QUALITY FILE %s WAS FOUND. REMAINING INPUT QUALITIES SET TO %d.", 
	    qf_name, parameters->default_qual);
    our_free(qf_name);
  }
/* rescale quality if necessary (done only if 0 < max_qual <= 20; rescaled values have
 maximum of 30 */
/*
  rescale_qual(db);
*/
  notify(" Done\n");
}

trim_qual(db, window, min_pos)
     Database *db;
  int window, min_pos;
{
  int entry1, length, cum_qual, j, k, j_min;
  char *qual;
  char *get_orig_qual();

  init_qual_hist();
  for (entry1 = db->first_entry; entry1 <= db->last_entry; entry1++) {
    qual = get_orig_qual(entry1);
    length = get_seq_length(entry1);
    cum_qual = 0;
    for (j = 0; j < window; j++) cum_qual += (qual[j] > 0);
    for (; cum_qual < min_pos && j < length; j++) 
      cum_qual += (qual[j] > 0) - (qual[j - window] > 0);
    for (k = 0; k < j - window; k++) qual[k] = 0;

    j_min = j;
    cum_qual = 0;
    for (j = length - 1; j >= length - window; j--) cum_qual += (qual[j] > 0);
    for (; cum_qual < min_pos && j > j_min; j--)
      cum_qual += (qual[j] > 0) - (qual[j + window] > 0);
    for (k = length - 1; k > j + window; k--) qual[k] = 0;
    incr_qual_hist(qual, length);
  }
  print_qual_hist("Trimmed quality");
}

trim_quals(db, qual_type)
     Database *db;
     int qual_type;
{
  Align_info *get_align_entry();
  char *get_adj_qual(), *get_orig_qual(), *get_id();
  Align_info *align_entry;
  char *qual;
  int i, length1, start, entry1;
  int cum, max_cum, best_start, best_end;

  for (entry1 = db->first_entry; entry1 <= db->last_entry; entry1++) {
    align_entry = get_align_entry(entry1);
    length1 = get_seq_length(entry1);
    qual = qual_type ? get_adj_qual(entry1) : get_orig_qual(entry1);

    if (!qual_type) {
      best_start = align_entry->qual_start;
      best_end = align_entry->qual_end;
    }
    else
      best_qual_seg(qual, length1, &best_start, &best_end);

    if (best_start > align_entry->first_start) best_start = align_entry->first_start;
    if (best_end < align_entry->last_end) best_end = align_entry->last_end;

    for (i = 0; i < best_start; i++) qual[i] = 0;
    for (i = best_end + 1; i < length1; i++) qual[i] = 0;

  }
}

trim_X_quals(db)
     Database *db;
{
  Align_info *get_align_entry();
  char *get_adj_qual(), *get_orig_qual(), *get_id(), *get_seq();
  Align_info *align_entry;
  char *qual, *seq;
  int i, length1, start, entry1;
  int start1, end1, start2, end2;

  for (entry1 = db->first_entry; entry1 <= db->last_entry; entry1++) {
    qual = get_orig_qual(entry1);
    seq = get_seq(entry1);
    for (i = 0; seq[i] && seq[i] != 'X'; i++);
    start1 = i;
    while (seq[i] == 'X') i++;
    end1 = i;
    for (; seq[i] && seq[i] != 'X'; i++);
    start2 = i;
    while (seq[i] == 'X') i++;
    end2 = i;
    if (start1 > parameters->vector_bound) start1 += 15;
    for (i = start1; i < end1 - 15; i++) qual[i] = 0;
    for (i = start2 + 15; i < end2; i++) qual[i] = 0;
  }
}

best_qual_seg(qual_array, length, add_start, add_end)
     char *qual_array;
     int length;
     int *add_start, *add_end;
{
  int i, cum, max_cum, start, q_start, q_end, q;

  q_start = length;
  q_end = -1;
  max_cum = cum = start = 0;
  for (i = 0; i < length; i++) {
    cum += scaled_err_probs[qual_array[i]];      /* qual_array[i] - trim_level; */
    if (cum > max_cum) {
      max_cum = cum;
      q_end = i;
      q_start = start;
    }
    else if (cum < 0) {
      cum = 0; 
      start = i + 1;
    }
  }
  *add_start = q_end > q_start + MIN_QUAL_SEG_SIZE ? q_start : length;
  *add_end = q_end > q_start + MIN_QUAL_SEG_SIZE ? q_end : -1;
}

find_trimmed_quals(align_entry, add_qual_start, add_qual_end)
     Align_info *align_entry;
     int *add_qual_start, *add_qual_end;
{
  int n_x, qual_start, qual_end;
  char *seq;
  char *get_seq();

  qual_start = align_entry->qual_start;
  qual_end = align_entry->qual_end;
  seq = get_seq(align_entry->seq_entry);

  for (n_x = 0; qual_end >= 0 && n_x < 10; qual_end--) 
    n_x += seq[qual_end] != 'X';
  if (qual_end >= 0) qual_end += 10;
  
  if (qual_start < align_entry->last_vec + 1) 
    qual_start = align_entry->last_vec + 1;

  *add_qual_end = qual_end;
  *add_qual_start = qual_start;
}

rescale_qual(db)
     Database *db;
{
  int i, max_qual;
  double scale_param;

  max_qual = get_max_qual();
  if (!max_qual || max_qual > 20) return;
  scale_param = 30.0 / max_qual;

  for (i = 0; i < db->t_length; i++) 
    db->orig_qual_area[i] *= scale_param;

}

/* sets adj_qual equal to orig_qual */
init_adj_qual(db)
     Database *db;
{
  int i;

  for (i = 0; i < db->t_length; i++) 
    db->adj_qual_area[i] = db->orig_qual_area[i];
}


/* convert (mostly) mononucleotide, quality < qual_cut, ends of sequence to a single char 
(usually 'N'); also converts trim_start letters at beginning to 'N' */

convert_ends(db) 
     Database *db;
{   
  int min_score, penalty, qual_cut;
  int trim_start, j, max, cum, max_pos;
  char c, end, end_char;
  char test_chars[30];
  char *chars, *seq;
  int entry1, length, qual_start, qual_end;
  char *get_seq(), *get_id();
  Align_info *align_entry;
  Align_info *get_align_entry();
  
  notify("Testing ends ...");
  min_score = parameters->trim_score;
  penalty = parameters->trim_penalty;
  trim_start = parameters->trim_start;
  end_char = 'N';
  qual_cut = 20;
  strcpy(test_chars,"ACGT");

  printf("\n\nFollowing regions converted to %c's\n", end_char);

  for (entry1 = db->first_entry; entry1 <= db->last_entry; entry1++) {
    align_entry = get_align_entry(entry1);
    qual_start = align_entry->qual_start;
    qual_end = align_entry->qual_end;
/*
    qual = get_orig_qual(entry1);
*/
    seq = get_seq(entry1);
    length = get_seq_length(entry1);
    for (end = 0; end <= 1; end++) {
      for (chars = test_chars; *chars; chars++) {
	cum = max = 0;
	for (j = end ? 0 : length - 1; 
/*
	     (end ? j < length : j >= 0) && qual[j] <= qual_cut; 
*/
	     end ? j < qual_start : j > qual_end;
	     j += (end ? 1 : -1)) {
	  c = seq[j];
	  cum += (c == *chars) ? 1 : (c == 'N' || c == 'X' ? 0 : penalty);
	  if (cum >= max) {
	    max = cum;
	    max_pos = j;
	  }
	}
	if (max >= min_score) {
	  printf("\n%s %d-%d ", get_id(entry1), end ? 1 : max_pos + 1, 
		 end ? max_pos + 1 : length);
	  for (j = end ? 0 : max_pos; j <= (end ? max_pos : length - 1); j++) {
	    printf("%c", seq[j]);
	    seq[j] = end_char;
	  }
	}      
      } 
    }
    for (j = 0; j < trim_start && seq[j]; j++) seq[j] = end_char;
  }     
  notify(" Done\n");
}

transform_contig_quality(contig1)
  Contig *contig1;
{
  int j;

  for (j = 0; j < contig1->length; j++)
    contig1->adj_qual[j] = -qual_LLR[contig1->adj_qual[j]];
}

print_quality(label, db)
     char *label;
     Database *db;
{
  int entry1;
  char *get_adj_qual();

  notify("Printing qualities ...");
  init_qual_hist();
  for (entry1 = db->first_entry; entry1 <= db->last_entry; entry1++) {
    incr_qual_hist(get_adj_qual(entry1), get_seq_length(entry1));
  }
  print_qual_hist(label);
  notify(" Done\n");
}

print_qual_segments(qual, length, cut)
     /* signed */ char *qual;
     int length, cut;
{
  int i, j, n_segs, t_size;

  for (i = 0; i < length && !qual[i]; i++);
  for (j = length - 1; j > i && !qual[j]; j--);
  printf("\n\nInitial, terminal qual 0 segments: ");
  if (i) printf(" 1-%d,", i);
  else printf(" (None),");
  if (j < length - 1) printf(" %d-%d", j + 2, length);
  else printf(" (None)");

  printf("\n\nRegions of LLR- adjusted quality < %.1f:\n", cut / 10.0);
  for (i = n_segs = t_size = 0; i < length; i = j) {
    for (; i < length && (int)qual[i] >= cut; i++);
    for (j = i; j < length && (int)qual[j] < cut; j++);
    if (j == i + 1) printf("%d, ", j);
    else if (j > i) printf("%d-%d, ", i + 1, j);
    t_size += j - i;
    n_segs++;
    if (!(n_segs % 8)) printf("\n");
  } 
  if (!n_segs) printf(" (None).");
  else printf("\n\n%d regions, avg size %.1f, avg spacing %.1f",
	      n_segs, t_size / (float)n_segs, length / (float)n_segs);
}

static int qual_hist_c[257];
static int *qual_hist;
static int num_entries;

init_qual_hist()
{
  int i;

  qual_hist = qual_hist_c + 1;
  num_entries = 0;
  for (i = -1; i < 256; i++) qual_hist[i] = 0;
}

incr_qual_hist(qual, length)
     /* signed */ char *qual;
     int length;
{
  int i;

  num_entries++;
  for (i = 0; i < length && !qual[i]; i++)
    qual_hist[-1] += 1;
  for (length--; !qual[length] && length > i; length--)
    qual_hist[-1] += 1;
  for (; i <= length; i++)
    qual_hist[qual[i]] += 1;
}

get_max_qual()
{
  int i;

  for (i = 255; i >= -1; i--) if (qual_hist[i]) return i;
  return 0;
}

print_qual_hist(label)
     char *label;
{
  int i, cum;
  double mean, total, div, cum_errs;

  printf("\n\n%s (quality, n_residues, %%,  cum, cum %%,  cum expected errs):", label);
  total = qual_hist[-1] + qual_hist[0];
  mean = 0;
  for (i = 1; i < 256; i++) {
    if (qual_hist[i]) {
      total += qual_hist[i];
      mean += (double)i * qual_hist[i];
    }
  }
  div = total ? total : 1;
  cum_errs = 0;
  for (i = 255, cum = 0; i >= -1; i--) 
    if (qual_hist[i]) {
      cum += qual_hist[i];
      cum_errs += qual_hist[i] * (i > 0 ? pow(10.0, i * -.1) : 1);
      printf("\n%3d  %6d %5.1f  %6d %5.1f  %6.2f", 
	     i, qual_hist[i], 100.0 * qual_hist[i] / div, cum, 100.0 * cum / div, cum_errs);
    }
  printf("   (quality -1 = terminal quality 0)");
  printf("\n\nAvg. full length: %.1f, trimmed (qual > -1): %.1f\nAvg. quality: %.1f per base", 
	 total / num_entries, (total - qual_hist[-1]) / num_entries, mean / div); 
}

static int *qual;

set_int_qual(c_qual, length, reverse)
     /* signed */ char *c_qual;
     int length, reverse;
{
  int j;
  char *our_alloc();

  qual = (int *)our_alloc(length * sizeof(int));
  for (j = 0; j < length; j++) {
    qual[j] = c_qual[reverse ? length - 1 - j : j];
    if (qual[j] > MAX_QUAL) fatalError("Quality out of bounds");
  }
  for (j = 0; j < length && !qual[j]; j++) qual[j] = -1;
  for (j = length - 1; !qual[j]; j--) qual[j] = -1;
}

free_int_qual()
{
  if (qual) our_free(qual);
}
 
print_diffs(pair, length2)
     Aligned_pair *pair;
     int length2;
{
  char *diff, *diff1;
  int i, j, q, length, print_flag;
  char *seq;
  int site1, site2, type;
  int d_site1, d_site2;
  char *get_seq();

  seq = get_seq(pair->entry1);
  length = get_seq_length(pair->entry1);
  site1 = pair->start1 - 1;
  site2 = pair->start2 - 1;
  for (diff = pair->diffs; *(diff + 1); diff++) {
    site1 += diff_gap1(*diff);
    site2 += diff_gap2(*diff);
    if ((type = diff_type(*diff)) == 'M') continue;
    /* note reversal of type : so that refers to change in 1st sequence with respect to the 2d */
    print_flag = 0;
    q = qual[site1];
    diff1 = diff;
    if (type == 'D') {
      for (diff1 = diff + 1; diff_type(*diff1) == 'D' && !diff_gap2(*diff1); diff1++);
      i = diff1 - diff; 
      diff1--;
      if (i > 1) {
	print_flag = 1;
	printf("\n%s I-%d %5d  ", 
	  parameters->tags ? "DISCREPANCY  " : "",
	       i, site1 + 1);
	for (j = q = 0; j < i; j++) {
	  printf("%c", seq[site1 + j]);
	  if (qual[site1 + j] > q) q = qual[site1 + j];
	}
      }
    }  
    else if (type == 'I') {
      for (diff1 = diff + 1; diff_type(*diff1) == 'I' && !diff_gap1(*diff1); diff1++);
      i = diff1 - diff; 
      diff1--;
      if (i > 1) {
	print_flag = 1;
	printf("\n%s D-%d %5d  %c", 
	  parameters->tags ? "DISCREPANCY  " : "",
	       i, site1 + 1, seq[site1]);
      }
    }	  
    if (!print_flag)
      printf("\n%s %c   %5d  %c", 
	  parameters->tags ? "DISCREPANCY  " : "",
	     type == 'D' ? 'I' : (type == 'I' ? 'D' : type),
	     site1 + 1, seq[site1]);
    printf("(%d)  %5d  ", q, 
	   is_reverse(pair) ? length2 - site2 : site2 + 1);
    d_site1 = site1;
    d_site2 = site2;
    if (diff < diff1)
      for (diff++; diff <= diff1; diff++) {
	d_site1 += diff_gap1(*diff);
	d_site2 += diff_gap2(*diff);
      }
    for (i = site1 > 6 ? site1 - 6 : 0; i < length && i < d_site1 + 7; i++) 
      printf("%c", i < site1 || i > d_site1 ? tolower(seq[i]) : seq[i]);

    diff = diff1;
    site1 = d_site1;
    site2 = d_site2;
  }
}

static int *n_align, *n_unalign, *n_p_unalign, *n_subs, *n_dels, *n_ins, *n_Ns, *n_Xs; 

init_diff_hist()
{
  int i;
  char *our_alloc();

  n_Ns = 1 + (int *)our_alloc((MAX_QUAL + 2) * sizeof(int));
  n_Xs = 1 + (int *)our_alloc((MAX_QUAL + 2) * sizeof(int));
  n_subs = 1 + (int *)our_alloc((MAX_QUAL + 2) * sizeof(int));
  n_dels = 1 + (int *)our_alloc((MAX_QUAL + 2) * sizeof(int));
  n_ins = 1 + (int *)our_alloc((MAX_QUAL + 2) * sizeof(int));
  n_align = 1 + (int *)our_alloc((MAX_QUAL + 2) * sizeof(int));
  n_unalign = 1 + (int *)our_alloc((MAX_QUAL + 2) * sizeof(int));
  n_p_unalign = 1 + (int *)our_alloc((MAX_QUAL + 2) * sizeof(int));
  for (i = -1; i <= MAX_QUAL; i++)
    n_subs[i] = n_Ns[i] = n_Xs[i] = n_dels[i] = n_ins[i] 
      = n_align[i] = n_unalign[i] = n_p_unalign[i] = 0;
}
 
incr_diff_hist(seq, diff, length1, length2, start1, start2, end1, end2, reverse)
     char *seq;
     char *diff;
     int length1, length2, start1, start2, end1, end2, reverse;
{
  int j, q, q1, offset;
  char *diff1;
  char c;
  int site1, site2, type;

  if (reverse) {
    site2 = start1 - 1;
    site1 = start2 - 1;
  }
  else {
    site1 = start1 - 1;
    site2 = start2 - 1;
  }
  for (; *(diff + 1); diff++) {
    site1 += diff_gap1(*diff);
    site2 += diff_gap2(*diff);
    if ((type = diff_type(*diff)) == 'M') continue;
    q = qual[reverse ? site2 : site1];
/*
    if (reverse && q >= parameters->qual_show) 
      printf("\n%d: quality %d", site1, q);
*/
    if (type == 'S') {
      c = seq[reverse ? site2 : site1];
      if (c == 'N') n_Ns[q] += 1;
      else if (c == 'X') n_Xs[q] += 1;
      else n_subs[q] += 1; 
    }
      
    else if (type == 'D') {
      for (diff1 = diff + 1; diff_type(*diff1) == 'D' && !diff_gap2(*diff1); diff1++)
	if ((q1 = qual[reverse ? site2 : site1 + (diff1 - diff)]) > q) q = q1;
      n_ins[q] += 1;
      site1 += (diff1 - diff) - 1;
      diff = diff1 - 1;
    } 
    /* note reversal of type : so that refers to change
       if 1st sequence with respect to the 2d */
      
    else if (type == 'I') {
      for (diff1 = diff + 1; diff_type(*diff1) == 'I' && !diff_gap1(*diff1); diff1++)
	if ((q1 = qual[reverse ? site2 + (diff1 - diff) : site1]) > q) q = q1;
      n_dels[q] += 1;
      site2 += (diff1 - diff) - 1;
      diff = diff1 - 1;
    }
  }

  offset = start2 - start1;
  for (j = 0; j < start1; j++) {
    if (qual[j] > MAX_QUAL || qual[j] < -1) {
      fprintf(stderr, "\n%d %d %d", j, qual[j], MAX_QUAL);
      fatalError("start1");
    }
    n_unalign[qual[j]] += 1;
    if (j + offset >= 0) n_p_unalign[qual[j]] += 1;
  }
  
  offset = end2 - end1;
  for (j = end1 + 1; j < length1; j++) {
    n_unalign[qual[j]] += 1;
    if (j + offset < length2) n_p_unalign[qual[j]] += 1;
  }
  
  for (j = start1; j <= end1; j++) n_align[qual[j]] += 1;
}

print_diff_hist(label)
     char *label;
{
  int i, t_errs, cum_align, rcum_align, rcum_errs, cum_errs, t_n_align, t_n_errs;

  printf("\n\n%s\nQual algn  cum    rcum    (%%)    unalgn X    N  sub del ins  total (%%)   cum  rcum (%%)", label);
  for (t_n_errs = t_n_align = i = 0; i <= MAX_QUAL; i++) {
    t_n_align += n_align[i];
    t_n_errs += n_subs[i] + n_dels[i] + n_ins[i] + n_Ns[i];
  }

  cum_align = cum_errs = 0;
  for (i = MAX_QUAL; i >= -1; i--) {
    if (!n_align[i] && !n_unalign[i]) continue;
    rcum_align = t_n_align - cum_align;
    cum_align += n_align[i];
    rcum_errs = t_n_errs - cum_errs;
    t_errs = n_subs[i] + n_dels[i] + n_ins[i] + n_Ns[i];
    cum_errs += t_errs;
    printf("\n%2d %6d %6d %6d (%6.2f)  %4d %2d   %2d %3d %3d %3d   %3d (%.2f)  %3d  %3d (%.2f)", 
	   i, /* n_align[i] + n_unalign[i], */
/*	   length ? (100.0 * (n_align[i] + n_unalign[i])) / length : 0.0, */
	   n_align[i], cum_align, 
/*	   t_n_align && i >= 0 ? 100.0 * cum_align / t_n_align : 0.0, */
	   rcum_align,
	   t_n_align && i >= 0 ? 100.0 * rcum_align / t_n_align : 0.0,
	   n_p_unalign[i], 
	   n_Xs[i], n_Ns[i], n_subs[i], n_dels[i], n_ins[i],
	   t_errs, 
	   n_align[i] ? (100.0 * t_errs) / n_align[i] : 0.0, 
	   cum_errs, 
/*	   cum_align && i >= 0 ? (100.0 * cum_errs) / cum_align : 0.0, */
	   rcum_errs, 
	   rcum_align && i >= 0 ? (100.0 * rcum_errs) / rcum_align : 0.0);
  }
  printf("\n");
}

free_diff_hist()
{
  our_free(n_Ns - 1);
  our_free(n_Xs - 1);
  our_free(n_subs - 1);
  our_free(n_dels - 1);
  our_free(n_ins - 1);
  our_free(n_align - 1);
  our_free(n_unalign - 1);
  our_free(n_p_unalign - 1);
}
