/***********************************************************************\
 *                                                                     * 
 *                           PROJECT   SSAHA                           *
 *                                                                     * 
 *      Sequence Search and Alignment by the Hashing Algorithm         *
 *                                                                     *
 *                                By                                   *
 *                                                                     *
 *                    Zemin Ning & James C. Mullikin                   *
 *                                                                     *
 *          Copyright (C) 1999-2004 by Genome Research Limited         *
 *                       All rights reserved                           *
 *                                                                     *
 *---------------------------------------------------------------------*
 #######################################################################
 #                                                                     #
 #             <------   LICENCE NOTICE   ------>                      #
 #                                                                     #
 # This is a licensed software developed by Genome Research Limited    #
 # (GRL) for genomic sequence assembling. For both commercial and non- # 
 # commercial purposes, a licence must be obtained from GRL before     #
 # use. Under no circumstances, the users should copy, modify and      #
 # redistribut the software as a whole or even part of it without      #
 # permission from GRL. For more information about the software and    #
 # its ducumentation particularly, please contact either one of the    # 
 # authors or GRL.                                                     #
 #######################################################################
 *---------------------------------------------------------------------*/

/* Flag options for the use of ssaha         

   The simplest use of the code is:

   -------------------------- 
   ssaha2 queryfile subjectfile
   --------------------------

/****************************************************************************/

 
#include <math.h>
#include <values.h>
#include <stdio.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <dirent.h>
#include <string.h>
#include <ctype.h>
#include "snp.h"
#include <unistd.h>
#include <sys/types.h> 
#include <sys/socket.h>

#define GT '>'
#define GT4 (((((GT<<8)+GT)<<8)+GT)<<8)+GT

#define MAX_DEPTH 50
#define MAX_PROJECTS 60
#define MAX_STRLEN 10240
#define MAX_READLENGTH 4096
#define MAX_FNAMELEN 512
#define ENDS_EXTRA 0
#define PADCHAR '-'
static char paths[MAX_PROJECTS][MAX_FNAMELEN];
static int nPaths;
#define MAX_N_NAM 300000 
#define MAX_N_BRG 500000 
#define MAX_N_ROW 40000 
#define Max_N_NameBase 100
#define Max_N_Pair 100
static int *sm,*head,*list,*readlength;
static int *sr,*sr_head,*sr_list;
static int *readpair_f,*readpair_b;
static int64_t n_Entry;
static int64_t MAX_S_LEN,MAX_R_LEN;

/* SSAS default parameters   */
static int K_LEN=12;
static int IMOD=0;
static int ISUB=0;
static int NCUT=200;
static int NHIT=10;
static int NBAT=9000000;
static int k_skip=12;
static int min_score=30;
static int all_out=0;
static int num_depth=50;
static int qvalue=0;
static int n_shift=3;
static int info_out=0;
static int selfcom_flag=0;
static int sense_flag=0;
static int align_flag=1;
static int tags_flag=0;
static int align_mem=2000;
static int client_flag=0;
static int max_len=1;
static int maxq_len=1;
static int maxs_len=1;
static int x_len=5000000;
static int y_len=50000;
static int est_flag=0;
static int n_start=0;
static int n_end=0;
static int LISTENQ=1024;
static int edge_len=100;
static int edge_set=0;
static int tidy_flag=0;
static float identy=50.0;
static int nfm=4000000;
static int debug_flag=0;
static int best_flag=0;
static int run_info=0;

typedef struct
{
       int foffset;
       int fsindex;
} SIO;

typedef struct
{
	int  index;
	int  length;
	char *data;
} subreads;

static SIO *fdata,*fdataRC;
static int *qinfo;
static int *qinforc;
static int *sinfo;
static int *sinforc;
static int reads[MAX_N_BRG][10];
static int reads2[MAX_N_BRG][10];

typedef struct
{
        char *name;
        char *name2;
        char *path;
        char *SCFname;
        int  length;
        char *data;
        char *qual;
        int  finished;
} fasta;

fasta *seqq,*seqs,*sub;
char *quelong;
char *sublong;

void error(char *msg)
{
    perror(msg);
    exit(1);
}

pid_t Fork(void)
{
  pid_t pid;
  if ( (pid = fork()) == -1) printf("fork error\n");
  return(pid);
}

void Close(int fd)
{  
  if(close(fd) == -1) 
  {
    printf("close error\n");
  }
}


int countQthresh (fasta *seq, int qthresh)
{
        int j;
        fasta *seqp;
        char *b;
        int n = 0;
        seqp = seq;
        if(seqp->finished) return(seqp->length);
        b = (char*)seqp->qual;
        for(j=seqp->length;--j>=0;b++)
                if(*b >= qthresh)
                        n++;
        return(n);
}

int duplicateRead(fasta *seq, fasta *dseq)
{
        int i;
        int size=((seq->data != 0)+(seq->qual != 0))*seq->length;
        char *tp,*dp;

        dseq->length = 0;
        dseq->data = NULL;
        if(size <= 0 || seq->data == NULL)
                return(1);

        if((dseq->data = (char *)malloc((size + 10*strlen(seq->name) + 1 )* sizeof(char))) == NULL)
        {
                printf("ERROR duplicateRead: malloc\n");
                return(1);
        }

        dseq->name = dseq->data+size;
        strcpy(dseq->name,seq->name);
        dseq->path = seq->path;
        dseq->SCFname = NULL;
        if(seq->SCFname != NULL)
        {
                dseq->SCFname = dseq->data+size+strlen(seq->name) + 1;
                strcpy(dseq->SCFname,seq->SCFname);
        }
        dseq->length = seq->length;
        if(seq->qual != 0)
                dseq->qual = dseq->data+dseq->length;

        dp = dseq->data;
        tp = seq->data;
        for(i=seq->length;--i>=0;)
                *dp++ = *tp++;

        if(seq->qual != 0)
        {
                dp = dseq->qual;
                tp = seq->qual;
                for(i=seq->length;--i>=0;)
                        *dp++ = *tp++;
        }
       
        return(1);

}


fasta *decodeFasta (char *fname, int *nContigs, int64_t *tB, int qThresh)
{
        FILE *fil;
        char *data = NULL, *dp, *cp;
        int64_t dataSize, totalbases=0, totalQbases=0;
        int64_t i;
        int ret;
        int nSeg;
        fasta *seg,*segp;
        int isfastq = 0;

        if ((fil = fopen(fname,"r")) == NULL)
        {
                printf("error decodeFastq: %s not found\n",fname);
                return (NULL) ;
        }

        fseek( fil, 0, SEEK_END);
        dataSize = ftell( fil);
        rewind( fil);

        if(dataSize == 0)
        {
                printf("error decodeFastq: no data in %s\n",fname);
                fclose(fil);
                return(NULL);
        }

        if((data = (char *) malloc( (dataSize+1) * sizeof(char) )) == NULL)
        {
                printf("error decodeFastq: malloc error\n");
                fclose(fil);
                return(NULL);
        }

        for(i=0;i<dataSize;i+=1024*1000000){
                int64_t size = 1024*1000000;
                if(dataSize-i < size)
                        size = dataSize-i;
//                printf("1024*1000000 chunck %d\n",i);
                if(fread(data+i,sizeof(char), size, fil) != size)
                {
                        printf("error decodeFastq: %s is too small\n",fname);
                        fclose(fil);
                        return (NULL) ;
                }
        }

        data[dataSize] = 0;
        nSeg = 0;
        dp = data;
        ret = 1;
        //printf("dataSize = %ld\n",dataSize);
        for(i=dataSize;--i>=0L;dp++)
        {
                if(*dp == '\n')
                {
                        ret = 1;
                        continue;
                }
                if(ret && (*dp == '@' || *dp == '>'))
                {
//                      printf("%19ld %6d %70.70s\n",dataSize-i,nSeg,dp);
                        nSeg++;
                }
                ret = 0;
        }
        //printf("i = %ld, dp = %ld\n",i,dp-data);

        if(nSeg == 0)
        {
                printf("error decodeFastq: no segments found\n");
                fclose(fil);
                return(NULL);
        }

        if((seg = (fasta *) calloc(nSeg, sizeof(fasta))) == NULL)
        {
                printf("error decodeFastq: calloc error\n");
                fclose(fil);
                return(NULL);
        }

        cp = dp = data;
        segp = seg;
        while(dp < data+dataSize)
        {
                if(*dp != '@' && *dp != '>')
                {
                        dp++;
                        continue;
                } else if ( dp > data && dp[-1] != '\n' ) {
                        dp++;
                        continue;
                } else {
                        int tmp;
                        int k;
                        if(*dp == '@') isfastq = 0;
                        if(*dp == '>') isfastq = 1;
                        dp++;
                        segp->name = (char *)(int64_t)(cp-data);
                        while((*cp++ = tmp = *dp++) != ' ' && tmp != '\n');
                        cp[-1] = 0;
                        if(tmp != '\n')
                        {
                                segp->name2 = (char *)(int64_t)(cp-data);
                                while((*cp++ = *dp++) != '\n');
                                cp[-1] = 0;
                        }
                        else
                                segp->name2 = NULL;
                        segp->data = (char *)(int64_t)(cp-data);
                        for(k=0;(tmp = *dp) && !((tmp == '+' || tmp == '>') && dp[-1] == '\n');dp++)
                        {
                                if(tmp == '\n')
                                        continue;
                                *cp++ = tmp;
                                k++;
                        }
                        segp->length = k;
                        if(isfastq) {
                                segp->finished = 1;
                                segp->qual = NULL;
                        } else {
                          segp->finished = 0;
                          k = 0;
                          if(tmp == '+')
                          {
                                int cp_set = 0;
                                int comp = 1;
                                while(*dp++ != '\n');
                                for(i=0;i<20;i++)
                                {
                                        if(dp[i] == 0)
                                                break;
                                        if(dp[i] == ' ')
                                        {
                                                comp = 0;
                                                break;
                                        }
                                }
                                segp->qual = (char *)(int64_t)(cp-data);
                                *cp = 0;
                                if(comp) for(k=0;(tmp = *dp) && !(tmp == '@' && dp[-1] == '\n');dp++)
                                {
                                        if(tmp == '\n')
                                                continue;
                                        *cp++ = tmp-041;
                                        k++;
                                }
                                else for(k=0;(tmp = *dp) && !(tmp == '@' && dp[-1] == '\n');dp++)
                                {
                                        if(tmp >= '0' && tmp <= '9')
                                        {
                                                if(*cp > 9) continue;
                                                *cp *= 10;
                                                *cp += tmp-'0';
                                                cp_set = 1;
                                                continue;
                                        }
                                        if(cp_set)
                                        {
                                                cp++;
                                                *cp = 0;
                                                cp_set = 0;
                                                k++;
                                        }
                                }
                          }
                          if(segp->length != k)
                          {
                            printf("Warning: %-20.20s (%ld) has %d bases an d %d quals, skipping entry\n",data+((int64_t)segp->name),strlen(data+((int64_t )segp->name)),segp->length,k);
                            cp = data+((int64_t)segp->name);
                            continue;
                          }
                        }
                        segp++;
                        if(segp-seg > nSeg)
                                break;
                }
        }
        nSeg = segp-seg;
        dataSize = cp - data;
//      printf("data memloc = %ld\n",data);
//      data = realloc(data,dataSize);
//      printf("data memloc = %ld\n",data);
        for(i=nSeg,segp=seg;--i>=0;segp++)
        {
                segp->name = data+((int64_t)segp->name);
                if(segp->name2)
                        segp->name2 = data+((int64_t)segp->name2);
                segp->data = data+((int64_t)segp->data);
                segp->qual = data+((int64_t)segp->qual);
                if(segp->finished) segp->qual = NULL;
                //printf("segment %s length %d\n",segp->name,segp->length);
                totalbases+=segp->length;
                totalQbases += countQthresh(segp,qThresh);
        }
        if(run_info == 1)
          printf("Total memory = %ld, total bases = %ld, Qbases(%d) = %ld, segments = %d\n",dataSize,totalbases,qThresh,totalQbases,nSeg);
        *nContigs = nSeg;
        *tB = totalbases;
        fclose(fil);
        return(seg);
}


/*   set all the base letters to low case  */
int fastaLC (fasta *seg, int nSeg)
{
	int i,j;
	fasta *segp;
	segp = seg;
	for(i=0;i<nSeg;i++,segp++)
	{
		char *b;
		b = (char*)segp->data;
/*              if(*b == tolower(*b))
			continue;                 */
		for(j=segp->length;--j>=0;b++)
			*b = tolower(*b);
	}
        return(1);
}


#define MAX_LINES 10000
#define MAX_EXPSIZE 100000

int setPath ( char *fullname, char **path )
{
	int len,i;
	len = strlen(fullname);
	while(--len >= 0 && fullname[len] != '/');
	*path = NULL;
	if(len >= 0)
	{
		len++;
		for(i=0;i<nPaths;i++)
			if(strncmp(fullname,paths[i],len) == 0)
				break;
		if(i == nPaths)
		{
			if(nPaths < MAX_PROJECTS)
			{
				strncpy(paths[nPaths],fullname,len);
				paths[nPaths][len] = 0;
				nPaths++;
			}
			else
			{
				printf("error setPath: max paths exceeded\n");
				exit(1);
			}
		}
		*path = paths[i];
	}
	return(0);
}
	
static char rc_char[3000];

int Reverse_Complement(fasta *seq, int seqdex)
{
	int i;
	char *tp,*dp;
        fasta *seqrc;

        seqrc=seq+seqdex;
        dp=rc_char;	
	tp = (seqrc->data)+(seqrc->length);
	for(i=seqrc->length;--i>=0;)
	{
		int tmp = *--tp;
		if     (tmp == 't') *dp++ = 'a';
		else if(tmp == 'g') *dp++ = 'c';
		else if(tmp == 'c') *dp++ = 'g';
		else if(tmp == 'a') *dp++ = 't';
		else                *dp++ = tmp;
	}
	return(0);
}


int reverseComplement(fasta *seq, fasta *rseq)
{
	int i;
	int size=((seq->data != 0)+(seq->qual != 0))*seq->length;
	char *tp,*dp;

	rseq->length = 0;
	if(size <= 0 || seq->data == NULL)
		return(1);

	if((rseq->data = (char *)malloc((size+strlen(seq->name)+1) * sizeof(char))) == NULL)
	{
		printf("ERROR reverseComplement: malloc\n");
		return(1);
	}

	rseq->name = rseq->data+size;
	strcpy(rseq->name,seq->name);
	rseq->path = seq->path;
	rseq->length = seq->length;
	if(seq->qual != 0)
		rseq->qual = rseq->data+rseq->length;
	
	dp = rseq->data;
	tp = seq->data+seq->length;
	for(i=seq->length;--i>=0;)
	{
		int tmp = *--tp;
		if     (tmp == 't') *dp++ = 'a';
		else if(tmp == 'g') *dp++ = 'c';
		else if(tmp == 'c') *dp++ = 'g';
		else if(tmp == 'a') *dp++ = 't';
		else                *dp++ = tmp;
	}
	if(seq->qual != 0)
	{
		dp = rseq->qual;
		tp = seq->qual+seq->length;
		for(i=seq->length;--i>=0;)
			*dp++ = *--tp;
	}
	return(0);
}


int main(int argc, char **argv)
{
    FILE *fpFasta,*fpSM;
    char NameFasta[50],NameQuality[50],FastaFile[50],cname[100]; 
    int i,i_Fasta,nSeq,sizesm,args;
    pid_t pid;
    fasta *seq,*seqp;
    DIR   *dp;
    struct dirent *dep;
    int readDepth[MAX_DEPTH],N_Fasta;
    void decodeReadpair(int nSeq);
    void HashFasta(fasta *seq, int nSeq, int i_Fasta, char NameFasta[50]);
    void HashQuality(fasta *seq, int nSeq, int i_Fasta, char NameQuality[50]);
    void Carry_Search(char cname[100]);
    void Trim_TupleRepeat(fasta *seq,int nSeq);
    void Trim_LowQuality_1(fasta *seq,int nSeq);
    void Trim_LowQuality_2(fasta *seq,int nSeq);
    void Trim_LowQuality_3(fasta *seq,int nSeq);
    void Trim_LowQuality_4(fasta *seq,int nSeq);
    void Exlude_WrongPairs(fasta *seq,int nSeq);
    void information();
    void memory_setup();
    int64_t tBase=0;
    int64_t MAX_S_LEN2;
    int skip_flag=0;

    for(i=0;i<MAX_DEPTH;i++) readDepth[i] = 0;
    if(argc < 3)
    {
      printf("USAGE\n");
      printf("\n");
      printf(" %s <query> <subject> <-seeds 10> <-kmer 12> <-cut 200> <-align 1> <-score 30> <-depth 50>\n",argv[0]);
      printf("\n");
      information();
      exit(1);
    }

    nSeq=0;
    args=3;
    for(i=args;i<argc;i++)
    {
       if(!strcmp(argv[i],"-mod"))
         sscanf(argv[++i],"%d",&IMOD); 
       else if(!strcmp(argv[i],"-sub"))
         sscanf(argv[++i],"%d",&ISUB); 
       else if(!strcmp(argv[i],"-cut"))
         sscanf(argv[++i],"%d",&NCUT); 
       else if(!strcmp(argv[i],"-hit"))
         sscanf(argv[++i],"%d",&NHIT); 
       else if(!strcmp(argv[i],"-seeds"))
         sscanf(argv[++i],"%d",&NHIT); 
       else if(!strcmp(argv[i],"-kmer"))
         sscanf(argv[++i],"%d",&K_LEN);
       else if(!strcmp(argv[i],"-len"))
         sscanf(argv[++i],"%d",&K_LEN);
       else if(!strcmp(argv[i],"-score"))
         sscanf(argv[++i],"%d",&min_score);
       else if(!strcmp(argv[i],"-bat"))
         sscanf(argv[++i],"%d",&NBAT);
       else if(!strcmp(argv[i],"-skip"))
       {
         sscanf(argv[++i],"%d",&k_skip);
         skip_flag=1;
       }
       else if(!strcmp(argv[i],"-allout"))
         sscanf(argv[++i],"%d",&all_out);
       else if(!strcmp(argv[i],"-depth"))
         sscanf(argv[++i],"%d",&num_depth);
       else if(!strcmp(argv[i],"-memory"))
         sscanf(argv[++i],"%d",&align_mem);
       else if(!strcmp(argv[i],"-edge"))
         sscanf(argv[++i],"%d",&edge_set);
       else if(!strcmp(argv[i],"-best"))
         sscanf(argv[++i],"%d",&best_flag);
       else if(!strcmp(argv[i],"-shift"))
         sscanf(argv[++i],"%d",&n_shift);
       else if(!strcmp(argv[i],"-info"))
         sscanf(argv[++i],"%d",&info_out);
       else if(!strcmp(argv[i],"-align"))
         sscanf(argv[++i],"%d",&align_flag);
       else if(!strcmp(argv[i],"-sense"))
         sscanf(argv[++i],"%d",&sense_flag);
       else if(!strcmp(argv[i],"-debug"))
         sscanf(argv[++i],"%d",&debug_flag);
       else if(!strcmp(argv[i],"-tags"))
         sscanf(argv[++i],"%d",&tags_flag);
       else if(!strcmp(argv[i],"-est"))
         sscanf(argv[++i],"%d",&est_flag);
       else if(!strcmp(argv[i],"-xlen"))
         sscanf(argv[++i],"%d",&x_len);
       else if(!strcmp(argv[i],"-start"))
         sscanf(argv[++i],"%d",&n_start);
       else if(!strcmp(argv[i],"-end"))
         sscanf(argv[++i],"%d",&n_end);
       else if(!strcmp(argv[i],"-client"))
         sscanf(argv[++i],"%d",&client_flag);
       else if(!strcmp(argv[i],"-idt"))
         sscanf(argv[++i],"%f",&identy);
       else if(!strcmp(argv[i],"-array"))
       {
         sscanf(argv[++i],"%d",&nfm);
         nfm = nfm*1000000;
       }
       else if(!strcmp(argv[i],"-tidy"))
         sscanf(argv[++i],"%d",&tidy_flag);
       else if(!strcmp(argv[i],"-runinfo"))
         sscanf(argv[++i],"%d",&run_info);
       else
       {
         printf("input parameter error\n");
         exit(1);
       }
    }

    if(run_info == 1)
    {
      fflush(stdout);
      system("ps aux | grep ssaha2; date");
    }
 
    n_Entry=1<<(K_LEN<<1);
    if(skip_flag==0)
      k_skip=K_LEN;

/*  allocate memory for arrays HEAD and LIST   */
    if((head = (int *)calloc((n_Entry),sizeof(int))) == NULL)
    {
      printf("ERROR ssaha: calloc - head\n");
      exit(1);
    }
    if((list = (int *)calloc(n_Entry,sizeof(int))) == NULL)
    {
      printf("ERROR ssaha: calloc - list\n");
      exit(1);
    }
    memory_setup();
    if(strcmp(argv[1],argv[2])==0)
      selfcom_flag=1;
/*--------------------------------------------------------------------*/
/*  Read both files of subject and query and generate a so-called
    sequence matrix SM using the hashing method, search the queryfile
    and print out the searched results                                */
/*--------------------------------------------------------------------*/
/*  read the subjectfile and hash it. But NOT Write output files !!!  */
/*--------------------------------------------------------------------*/
    if(est_flag==0)
    {
      x_len=10000;
      y_len=20;
    }
    else
    {
      x_len=5000000;
      y_len=50000;
    }
    sprintf(NameFasta,"%s",argv[2]);
    sub = decodeFasta(argv[2],&nSeq,&tBase,qvalue);
    i_Fasta=1;
/*  define the sequence matrix SM     */
    MAX_S_LEN=(tBase/k_skip)+100;
    MAX_S_LEN2=MAX_S_LEN<<1;
    if((sm = (int *)calloc(MAX_S_LEN2,sizeof(int))) == NULL)
    {
      printf("ERROR HashFasta: calloc - sm\n");
      exit(1);
    }

/*  define the array readlength     */
    if((readlength = (int *)calloc(nSeq,sizeof(int))) == NULL)
    {
      printf("ERROR HashFasta: calloc - readlength\n");
      exit(1);
    }

    fastaLC(sub,nSeq);
    HashFasta(sub,nSeq,i_Fasta,NameFasta);

    if(strncmp(argv[1],"client",6)==0)
      client_flag=1;
    if(client_flag==0)
    {
      strcpy(cname,argv[1]);
      Carry_Search(cname);
    }
    else
    {
      int sockfd, newsockfd, portno,clilen,out_flag;
      char buffer[256];
      struct sockaddr_in serv_addr, cli_addr;
      int n;

      sockfd = socket(AF_INET, SOCK_STREAM, 0);
      if (sockfd < 0)
         error("ERROR opening socket");
      bzero((char *) &serv_addr, sizeof(serv_addr));
      portno = 777777;
      serv_addr.sin_family = AF_INET;
      serv_addr.sin_addr.s_addr = INADDR_ANY;
      serv_addr.sin_port = htons(portno);
      if (bind(sockfd, (struct sockaddr *) &serv_addr,
              sizeof(serv_addr)) < 0)
              error("ERROR on binding");
      listen(sockfd,LISTENQ);
      out_flag=0;
      printf("new process 1\n");
      for(;;)
      {
         clilen = sizeof(cli_addr);
         if((newsockfd = accept(sockfd, (struct sockaddr *) &cli_addr, &clilen))< 0)
           error("ERROR on accept");
         if( (pid = Fork()) == 0) 
         {
           Close(sockfd);

           printf("new process 2\n");
           while(out_flag==0)
           {
             bzero(buffer,256);
             bzero(cname,100);
             n = read(newsockfd,buffer,255);
             if (n < 0) error("ERROR reading from socket");
             printf("Here is the message: %s %d\n",buffer,strlen(buffer));
             n = write(newsockfd,"I got your message",18);
             if (n < 0) error("ERROR writing to socket");
             if(strncmp(buffer,"stop",4)==0)
               out_flag=1;
             strncpy(cname,buffer,strlen(buffer)-1);
             if(out_flag==0)
               Carry_Search(cname);
           }
           printf("I want to stop1\n");
//           Close(newsockfd);
//           exit(0);
         }
         printf("I want to stop: %d\n",out_flag);
//         close(newsockfd);
//         exit(0);
         if(out_flag==1)
           exit(1);
      }
    }
    printf("Job finished for %d input sequences!\n",nSeq);
    return EXIT_SUCCESS;

}
/* end of the main */

/*   Subroutine to hash the DNA sequences into a matrix table      */
/* ==================================================================  */
void Carry_Search(char cname[100])
/* ==================================================================  */
{
     int i,nSeq;
     fasta *seq,*seqp;
     void Search_SM(fasta *seq,int nSeq);
     int64_t tBase;


     if((seq = decodeFasta(cname,&nSeq,&tBase,qvalue))==NULL)
       printf("wrong query file name!\n");
     else
     {
       fastaLC(seq,nSeq);
       maxq_len=1;
       seqp=seq;
       for(i=0;i<nSeq;i++,seqp++)
       {
          if(seqp->length>maxq_len)
          maxq_len=seqp->length;
       } 
       Search_SM(seq,nSeq);
     }
}

/*   Subroutine to hash the DNA sequences into a matrix table      */
/* ==================================================================  */
void HashFasta(fasta *seq, int nSeq, int i_Fasta, char NameFasta[50])
/* ==================================================================  */
{
     int i,j,k=0,k_Base=K_LEN;
     int ps=0,ns=0,IntSeg=0,IntBase=0;
     int64_t dataSize;
     FILE *genfil;
     char *b,*st,*ch,*ed,*st2,*ed2;
     fasta *seqp;
     float ntotal;
     void ArraySort_String(int n,char **Pair_Name,int *brr);
     void Get_ReadPairs(int rd,int read_pair[Max_N_Pair],int *arr);
     int rd,read_pair[Max_N_Pair],i_pair;
     char **cmatrix(long nrl,long nrh,long ncl,long nch);

/*   sort all the names of genes or name entries   */
     seqp = seq;
     for(i=0;i<nSeq;i++,seqp++)
     {
        b = (char*)seqp->data;
        ns=((seqp->length)+k_skip-k_Base)/k_skip;
        readlength[i]=seqp->length;
        if(readlength[i]>maxs_len)
          maxs_len=readlength[i];
        for(j=0;j<ns;j++)
        {
           IntSeg=0;
           ch=b;
           for(k=k_Base;k>0;k--)
           {
              if     (*ch == 'a') IntBase=0;
              else if(*ch == 'c') IntBase=1;
              else if(*ch == 'g') IntBase=2;
              else if(*ch == 't') IntBase=3;
              else if(*ch == 'n') IntBase=1;
              else
              {
                IntBase=1;
        /*        printf("ERROR ssaha: data input %c %d\n",*ch,i);
                exit(1);   */
              }
              IntBase=IntBase<<((k-1)<<1);
              IntSeg=IntSeg+IntBase;
              ch++;
           }
           head[IntSeg]=head[IntSeg]+1;
           b=b+k_skip;
        }
     }

/*   accumulate the HEAD array to point the starting position of the entries */
     for(i=1;i<n_Entry;i++)
        head[i]=head[i]+head[i-1];

/*   hash all the strings with k_Base pairs into Matrix array sm   */
     seqp = seq;
     for(i=0;i<nSeq;i++,seqp++)
     {
        b = (char*)seqp->data;
        ns=((seqp->length)+k_skip-k_Base)/k_skip;
        for(j=0;j<ns;j++)
        {
           IntSeg=0;
           ch=b;
           for(k=k_Base;k>0;k--)
           {
              if     (*ch == 'a') IntBase=0;
              else if(*ch == 'c') IntBase=1;
              else if(*ch == 'g') IntBase=2;
              else if(*ch == 't') IntBase=3;
              else if(*ch == 'n') IntBase=1;
              else
              {
                IntBase=1;
        /*        printf("ERROR ssaha: data input %c\n",*ch);
                exit(1);    */
              }
              IntBase=IntBase<<((k-1)<<1);
              IntSeg=IntSeg+IntBase;
              ch++;
           }
           if(IntSeg!=0)
             ps=list[IntSeg]+head[IntSeg-1];
           else  
             ps=list[IntSeg];
           sm[ps]=i;
           sm[ps+MAX_S_LEN]=j*k_skip+1;
           list[IntSeg]=list[IntSeg]+1;
           b=b+k_skip;
        }
     }
}

static int n_OverlapReads;
// static fasta vect;

/*   Subroutine to search sequence while reporting matching positions    */
/* =============================== */
void Search_SM(fasta *seq,int nSeq)
/* =============================== */
{
     int i,j,k,g,m,k_Base=K_LEN,k_base1,bignum;
     int cm1[MAX_N_ROW],cm2[MAX_N_ROW],cm3[MAX_N_ROW];
     int SubBase[k_Base],SubBaseRC[k_Base],Base[4]={0,1,2,3};
     int IntSub[k_Base][4],IntSubRC[k_Base][4];
     int nline=0,ns=0,n_Base,jj,nsft,RePeat,NonMat,nsrc;
     int IntSeg,IntBase,IntSegRC,IntBaseRC,IntSegBfor,IntSegBforRC,IntA;
     int Match_flag,TR_flag,IntSegBfor2,IntSegBfor2RC;
     int IntSegBfor3,IntSegBfor4,IntSegBfor3RC,IntSegBfor4RC;
     int tandem,tandem_flag,n_tandem,n_tandem_rc;
     int ifm,ifmrc,n_freq,b_flag,n_hit,shift,s_length,q_length;
     int64_t MAX_S_LEN2;
     int offset=0,sizesm=0,batch,iseq,nbatch=0,wft,wrc;
     FILE *fpSM;
     char *b,*ch,*st,sqname[34],signRC;
     fasta *seqp;
     void ArraySort_Mix3(int n, int64_t *arr, int *brr, int *crr);
     void ArraySort_Int2(int n, int *arr, int *brr);
     void ArraySort2_Int2(int n, int *arr, int *brr);
     int  reverseComplement(fasta *seq, fasta *seqp);
     SIO *p,*prc;
     int *qqinfo,*qqinforc,*ssinfo,*ssinforc;
     int *mask_que,*mask_sub,*cross_parameters;
     int *hit_length,*hit_index;
     int **align;
     int64_t  *indexset,*indexsetRC,*ptr;
     int  **imatrix(int64_t nrl,int64_t nrh,int64_t ncl,int64_t nch);
     int read1st_flag=0,num_align;
     int cross_match(fasta *seq1,fasta *seq2,int *grr,char *e1r,char *e2r,int **align,int *hrr);

     
/*   setting initial values   */
     seqp = seq;
     seqq = seq;
     seqs = seq;
     iseq=0;
     bignum=1<<24;
     k_base1=(k_Base-1)<<1;
     MAX_S_LEN2=MAX_S_LEN<<1;

     if((mask_que = (int *)calloc(maxq_len,sizeof(int))) == NULL)
     {
       printf("ERROR HashFasta: calloc - mask_que\n");
       exit(1);
     }
     if((mask_sub = (int *)calloc(maxs_len,sizeof(int))) == NULL)
     {
       printf("ERROR HashFasta: calloc - mask_sub\n");
       exit(1);
     }

     if((cross_parameters = (int *)calloc(20,sizeof(int))) == NULL)
     {
       printf("ERROR HashFasta: calloc - cross_parameters\n");
       exit(1);
     }
     if((hit_length = (int *)calloc(MAX_N_BRG,sizeof(int))) == NULL)
     {
       printf("ERROR HashFasta: calloc - hit_length\n");
       exit(1);
     }
     align=imatrix(0,1000,0,20);
     if((hit_index = (int *)calloc(MAX_N_BRG,sizeof(int))) == NULL)
     {
       printf("ERROR HashFasta: calloc - hit_index\n");
       exit(1);
     }
     max_len= maxq_len>maxs_len ? maxq_len : maxs_len;
     if((quelong = (char *)calloc(max_len,sizeof(char))) == NULL)
     {
       printf("ERROR HashFasta: calloc - quelong\n");
       exit(1);
     }
     if((sublong = (char *)calloc(max_len,sizeof(char))) == NULL)
     {
       printf("ERROR HashFasta: calloc - sublong\n");
       exit(1);
     }
/*  allocate memory for structures fdata and fdataRC   */
/*    if((fdata = (SIO *) calloc(nfm,sizeof(SIO))) == NULL)
     {
       printf("ERROR ssaha: calloc - fdata\n");
       exit(1);
     }
     if((fdataRC = (SIO *) calloc(nfm,sizeof(SIO))) == NULL)
     {
       printf("ERROR ssaha: calloc - fdataRC\n");
       exit(1);
     }       */


     n_OverlapReads=0;
     s_length=0;
     ifm=0;
     ifmrc=0;
     shift=n_shift*k_Base;
     if(n_end==0)
       n_end = nSeq;
//     for(iseq=n_start;iseq<nSeq;iseq++) 
     for(iseq=n_start;iseq<n_end;iseq++) 
     {
       int kk,kc;
       seqp=seq+iseq;
       seqq=seqp;
       cross_parameters[8]=0;
       num_align=0;
       if(ifm>0)
         for(p=fdata;p<fdata+ifm;p++)
         {
            p->fsindex=0;
            p->foffset=0;
         }
       if(ifmrc>0)
         for(prc=fdataRC;prc<fdataRC+ifmrc;prc++)
         {
            prc->fsindex=0;
            prc->foffset=0;
         }

       b = ((char*)seqp->data);
       ns=(seqp->length)/k_Base;
       if(ns<=NBAT)
       {
         nbatch=1;
/*         ns=ns-1;   */
       }
       else
       {
         nbatch=ns/NBAT;
         ns=NBAT;
       }

       q_length=seqp->length;
       if(edge_set==0)
       {
         if(q_length<2000)
           edge_len=200;
         else
           edge_len=q_length/5;
         if(edge_len<200)
           edge_len=200;
         else if(edge_len>10000)
           edge_len=10000;

       }
       else
         edge_len=edge_set; 
       
       n_Base=q_length+1-k_Base;
       if(tidy_flag==0)
       {
         if(iseq!=n_start)
           printf("\n");
         printf("===================================================\n");
         printf("Matches For Query %d (%d bases): %s\n",iseq,q_length,seqp->name);
         printf("===================================================\n");
         printf("Score     Q_Name             S_Name            Q_Start    Q_End  S_Start    S_End Direction #Bases identity\n");         
       }
       for(batch=1;batch<=nbatch;batch++)
       {
        if(ns<10)              n_hit=0.6*ns;
        else if(ns>=10&&ns<20) n_hit=0.7*ns;
        else                   n_hit=NHIT;
//        else                   n_hit=0.05*ns;
        p=fdata;
        qqinfo=qinfo;
        ssinfo=sinfo;
        prc=fdataRC;
        qqinforc=qinforc;
        ssinforc=sinforc;
        IntSegBfor=-1;
        IntSegBforRC=-1;
        IntSegBfor2=-2;
        IntSegBfor2RC=-2;
        IntSegBfor3=-3;
        IntSegBfor3RC=-3;
        IntSegBfor4=-4;
        IntSegBfor4RC=-4;
        ifm=0;
        ifmrc=0;
        ch=b;
        tandem=0;
        n_tandem=0;
        n_tandem_rc=0;
        tandem_flag=0;

/*      get the entry integer       */
        for(i=0;i<n_Base;i++)
        {
              int n_Ns=0,err_flag=0;
              float ratio=0.0;
              IntSeg=0;
              IntSegRC=0;
              ch=b;
              jj=i%k_Base;
              for(k=k_Base;k>0;k--)
              {
                 if     (*ch == 'a') IntBase=0;
                 else if(*ch == 'c') IntBase=1;
                 else if(*ch == 'g') IntBase=2;
                 else if(*ch == 't') IntBase=3;
                 else if(*ch == '\0')
                 {
                   printf("Cannot find any match in the database! s2 %c\n",*ch);
                   exit(1);
                 }
                 else
                 {
                   IntBase=1;
                   n_Ns++;
                 }
                 IntBaseRC=(IntBase^3)<<((k_Base-k)<<1);
                 IntBase=IntBase<<((k-1)<<1);
                 SubBase[k_Base-k]=IntBase;
                 SubBaseRC[k_Base-k]=IntBaseRC;
                 IntSeg=IntSeg+IntBase;
                 IntSegRC=IntSegRC+IntBaseRC;
                 ch++;
              }
              ratio=n_Ns;
              ratio=n_Ns/k_Base;
              err_flag= ratio>0.7 ? 0 : 1;

/*            Obtain the integers when one substitution is allowed   */
              if(ISUB==1)
              {
                for(g=0;g<k_Base;g++)
                {
                   IntA=IntSeg-SubBase[g];
                   IntSub[g][0]=IntA; 
                   IntSub[g][1]=IntA+(Base[1]<<((k_Base-g-1)<<1)); 
                   IntSub[g][2]=IntA+(Base[2]<<((k_Base-g-1)<<1)); 
                   IntSub[g][3]=IntA+(Base[3]<<((k_Base-g-1)<<1));
                   IntA=IntSegRC-SubBaseRC[g];
                   IntSubRC[g][0]=IntA+(3<<((g-1)<<1)); 
                   IntSubRC[g][1]=IntA+(2<<((g-1)<<1)); 
                   IntSubRC[g][2]=IntA+(1<<((g-1)<<1)); 
                   IntSubRC[g][3]=IntA;
                }
              }

/*            FRONT: generate two arrays for frequence calculation    */
              RePeat=0;
              NonMat=0;
              for(g=0;g<k_Base;g++)
              {
                 for(m=0;m<4;m++)
                 {
                    if(ISUB==0||ns<=10)
                    {
                      m=4;g=k_Base;
                      nsft=IntSeg;
                    }
                    else
                    {
/*                    exclude the repeats when there is an exact match  */
                      if(IntSub[g][m]==IntSeg) 
                      { 
                         RePeat++;NonMat=0;
                      }
                      else                    
                         NonMat=1;
                      nsft=IntSub[g][m];
                    }
                    wft=head[nsft]-head[nsft-1];
                    TR_flag=(IntSeg!=IntSegBfor)&&(IntSeg!=IntSegBfor2)&&(IntSeg!=IntSegBfor3)&&(IntSeg!=IntSegBfor4);
                    Match_flag=err_flag;
                    if((wft!=0)&&(wft<=NCUT)&&(Match_flag)&&(TR_flag))
                    {
                      NonMat=0;
/*                    read data from the stored file: sm.out   */
                      if(IMOD==2||IMOD==4)
                      {
                        if((fpSM=fopen("sm.out","rb"))==NULL)
                        {
                          printf("Cannot open file: sm.out\n");
                          exit(0);
                        }
                        sizesm=wft;
                        offset=head[nsft-1]*sizeof(int);
                        fseek(fpSM,offset,0);
                        fread(cm1,sizeof(int),sizesm,fpSM);
                        rewind(fpSM);

                        offset=(head[nsft-1]+MAX_S_LEN)*sizeof(int);
                        fseek(fpSM,offset,0);
                        fread(cm2,sizeof(int),sizesm,fpSM);
                        rewind(fpSM);
                        
                        offset=(head[nsft-1]+MAX_S_LEN2)*sizeof(int);
                        fseek(fpSM,offset,0);
                        fread(cm3,sizeof(int),sizesm,fpSM);
                        rewind(fpSM);
                        fclose(fpSM);

/*                      generate the frequence array     */
                        offset=head[nsft-1];
                        for(k=0;k<wft;k++)
                        {
                           p->foffset=cm2[k]-i+bignum;
                           p->fsindex=cm1[k];
                           qinfo[ifm]=i;
                           sinfo[ifm]=cm2[k];
                           ifm++;  p++;
                           if(ifm>=nfm) 
                           {
                             printf("Frequence array is full \n");
                             break;
                           }
                        }
                      }
/*                    get data from matrix table SM    */
                      else
                      {
/*                      generate the frequence array     */
                        offset=head[nsft-1];
                        tandem=0;
                        if(nsft==0) offset=0;
                        for(k=0;k<wft;k++)
                        {
                           if(!tandem)
                           {
                             tandem_flag=(sm[offset+k+1]==sm[offset+k])&&(sm[offset+k+1+MAX_S_LEN]==(sm[offset+k+MAX_S_LEN]+k_Base));
                             if(tandem_flag)
                             {
                               p->foffset=sm[offset+k+MAX_S_LEN]-i+n_tandem*12+bignum;
                               n_tandem=1+(i/k_Base);
                             }
                             else
                               p->foffset=sm[offset+k+MAX_S_LEN]-i+bignum; 
                             p->fsindex=sm[offset+k];
                             s_length=readlength[p->fsindex];
                             qinfo[ifm]=i;
                             sinfo[ifm]=sm[offset+k+MAX_S_LEN]; 
//     if((i+1)%k_Base==0)
//     printf("------------------------------------------------------\n");
//     printf("link-ft: %d %d %d %d %d %d %d %d %d %d \n",ifm,i,k,wft,IntSeg,IntSegBfor,sm[offset+k+MAX_S_LEN],sm[offset+k],qinfo[ifm],(i+1)%12);
                             ifm++;  p++;
                             if(ifm>=nfm) 
                             {
                               printf("Frequence array is full \n");
                               break;
                             }
                           }
                           tandem=(sm[offset+k+1]==sm[offset+k])&&(sm[offset+k+1+MAX_S_LEN]==(sm[offset+k+MAX_S_LEN]+k_Base));
                        }
                      }
                    }
/*            new insert   */
                 }              
              }              

/*            Reverse complement: generate two arrays for frequence  */
              RePeat=0;
              NonMat=0;
              for(g=0;g<k_Base;g++)
              {
                 for(m=0;m<4;m++)
                 {
                    if(ISUB==0||ns<=10)
                    {
                      m=4;g=k_Base;
                      nsrc=IntSegRC;
                    }
                    else
                    {
/*                    exclude the repeats when there is an exact match  */
                      if(IntSubRC[g][m]==IntSegRC) 
                      { 
                         RePeat++;NonMat=0;
                      }
                      else                    
                         NonMat=1;
                      nsrc=IntSubRC[g][m];
                    }
                    wrc=head[nsrc]-head[nsrc-1];
                    TR_flag=(IntSegRC!=IntSegBforRC)&&(IntSegRC!=IntSegBfor2RC)&&(IntSegRC!=IntSegBfor3RC)&&(IntSegRC!=IntSegBfor4RC);
                    Match_flag=err_flag;
                    if((wrc!=0)&&(wrc<=NCUT)&&(Match_flag)&&(TR_flag))
                    {
                      NonMat=0;
/*                    read data from the stored file: sm.out   */
                      if(IMOD==2||IMOD==4)
                      {
                        if((fpSM=fopen("sm.out","rb"))==NULL)
                        {
                          printf("Cannot open file: sm.out\n");
                          exit(0);
                        }
                        sizesm=wrc;
                        offset=head[nsrc-1]*sizeof(int);
                        fseek(fpSM,offset,0);
                        fread(cm1,sizeof(int),sizesm,fpSM);
                        rewind(fpSM);

                        offset=(head[nsrc-1]+MAX_S_LEN)*sizeof(int);
                        fseek(fpSM,offset,0);
                        fread(cm2,sizeof(int),sizesm,fpSM);
                        rewind(fpSM);
                        fclose(fpSM);

/*                      generate the frequence array     */
                        for(k=0;k<wrc;k++)
                        {
                           prc->foffset=cm2[k]+i;
                           prc->fsindex=cm1[k];
                           qinforc[ifmrc]=i;
                           sinforc[ifmrc]=cm2[k];
                           ifmrc++;    prc++;
                           if(ifmrc>=nfm) 
                           {
                             printf("Frequence array is full\n");
                             break;
                           }
                        }
                      }
/*                    get data from matrix table SM    */
                      else
                      {
/*                      generate the frequence array     */
                        offset=head[nsrc-1];
                        tandem=0;
                        if(nsrc==0) offset=0;
                        for(k=0;k<wrc;k++)
                        {
                           if(!tandem)
                           {
                             tandem_flag=(sm[offset+k+1]==sm[offset+k])&&(sm[offset+k+1+MAX_S_LEN]==(sm[offset+k+MAX_S_LEN]+k_Base));
                             if(tandem_flag)
                             {
//                        if(sm[offset+k]!=quedex)
                               prc->foffset=sm[offset+k+MAX_S_LEN]+i-n_tandem_rc*k_Base;
                               n_tandem_rc=1+(i/k_Base);
                             }
                             else
                               prc->foffset=sm[offset+k+MAX_S_LEN]+i;
                             prc->fsindex=sm[offset+k];
                             qinforc[ifmrc]=i;
                             sinforc[ifmrc]=sm[offset+k+MAX_S_LEN];
                             s_length=readlength[prc->fsindex];
//     if((i+1)%k_Base==0)
//     printf("------------------------------------------------------\n");
//     printf("link-rc: %d %d %d %d %d %d %d %d %d \n",ifmrc,i,k,wrc,IntSegRC,(s_length-k_Base+2-sm[offset+k+MAX_S_LEN]),sm[offset+k],qinforc[ifm],(i+1)%12);
                             ifmrc++;    prc++;
                             if(ifmrc>=nfm) 
                             {
                               printf("Frequence array is full\n");
                               break;
                             }
                           }
                           tandem=(sm[offset+k+1]==sm[offset+k])&&(sm[offset+k+1+MAX_S_LEN]==(sm[offset+k+MAX_S_LEN]+k_Base));
                        }
                      }
                    }
/*            new insert   */
                 }              
              }              

              if(ifm>=nfm) break;
              if(ifmrc>=nfm) break;
              IntSegBfor4=IntSegBfor3;
              IntSegBfor4RC=IntSegBfor3RC;
              IntSegBfor3=IntSegBfor2;
              IntSegBfor3RC=IntSegBfor2RC;
              IntSegBfor2=IntSegBfor;
              IntSegBfor2RC=IntSegBforRC;
              IntSegBfor=IntSeg;
              IntSegBforRC=IntSegRC;
              if(i%k_Base==0||(i+1)%k_Base==0)
              {
                IntSegBfor4=-4;
                IntSegBfor4RC=-4;
                IntSegBfor3=-3;
                IntSegBfor3RC=-3;
                IntSegBfor2=-2;
                IntSegBfor2RC=-2;
              }
              b++;
        }



/*      search from the front direction    */ 
        p=fdata;
        qqinfo=qinfo;
        ssinfo=sinfo;
        indexset=(int64_t *)fdata;
        ptr=indexset;
        ArraySort_Mix3(ifm,indexset,qqinfo,ssinfo);
//        for(k=0;k<ifm;k++)
//           printf("fm: %d %ld %d %d %ld %ld %d %d\n",k,*(ptr+k),(p+k)->fsindex,(p+k)->foffset,ptr[k],ptr[k]-ptr[k-1],qinfo[k],ssinfo[k]);    
        n_OverlapReads=0;
        b_flag=0;
        for(k=0;k<ifm-1;k++)
        {
           if((ptr[k+1]-ptr[k]<=shift)&&((p+k+1)->foffset>0))
           {
             j=k+1;
             n_freq=1;
             while((j<ifm)&&(ptr[j]-ptr[k]<=shift))
             {
                j++;
                n_freq++;
             }
             if((n_freq>=n_hit)&&(n_OverlapReads<MAX_N_BRG))
             {
               qqinfo=qinfo;
               ssinfo=sinfo;
               ArraySort_Int2(j-k,qqinfo+k,ssinfo+k);
//        for(m=k;m<k+n_freq;m++)
//           printf("fm: %d %ld %d %d %ld %ld %d %d %d\n",m,*(ptr+m),(p+m)->fsindex,(p+m)->foffset-bignum,ptr[m],ptr[m]-ptr[m-1],qinfo[m],k,sinfo[k]);    
               reads[n_OverlapReads][0]=(p+k)->fsindex;
               reads[n_OverlapReads][1]=sinfo[k];
               reads[n_OverlapReads][2]=n_freq;
               reads[n_OverlapReads][3]=0;
               reads[n_OverlapReads][4]=qinfo[k]+1;
               reads[n_OverlapReads][5]=qinfo[j-1]+k_Base;
               reads[n_OverlapReads][6]=sinfo[j-1]+k_Base-1;
               reads[n_OverlapReads][7]=1;
               reads2[n_OverlapReads][1]=sinfo[k];
               reads2[n_OverlapReads][4]=qinfo[k]+1;
               reads2[n_OverlapReads][5]=qinfo[j-1]+k_Base;
               reads2[n_OverlapReads][6]=sinfo[j-1]+k_Base-1;
               n_OverlapReads++;
             }
             k=j;
           }
        }
 
/*      search from reverse complement    */
        p=fdataRC;  
        qqinforc=qinforc;
        ssinforc=sinforc;
        indexsetRC=(int64_t *)fdataRC;
        ptr=indexsetRC;
        ArraySort_Mix3(ifmrc,indexsetRC,qqinforc,ssinforc);
        b_flag=0;
//        for(k=0;k<ifmrc;k++)
//           printf("fmrc: %d %ld %d %d %ld %ld %d %d\n",k,*(ptr+k),(p+k)->fsindex,(p+k)->foffset,ptr[k],ptr[k]-ptr[k-1],qinforc[k],ssinforc[k]);    
        for(k=0;k<ifmrc-1;k++)
        {
           if((ptr[k+1]-ptr[k]<=shift)&&((p+k+1)->foffset>0))
           {
             j=k+1;
             n_freq=1;
             while((j<ifmrc)&&(ptr[j]-ptr[k]<=shift))
             {
                j++;
                n_freq++;
             }
             if((n_freq>=n_hit)&&(n_OverlapReads<MAX_N_BRG))
             {
               qqinforc=qinforc;
               ssinforc=sinforc;
               ArraySort_Int2(j-k,qqinforc+k,ssinforc+k);
               s_length=readlength[(p+k)->fsindex];
//               reads[n_OverlapReads][1]=s_length-sinforc[k]-k_Base+2;
               reads[n_OverlapReads][0]=(p+k)->fsindex;
               reads[n_OverlapReads][1]=sinforc[j-1];
               reads[n_OverlapReads][2]=n_freq;
               reads[n_OverlapReads][3]=1;
               reads[n_OverlapReads][4]=qinforc[k]+1;
               reads[n_OverlapReads][5]=qinforc[j-1]+k_Base;
               reads[n_OverlapReads][6]=sinforc[k]+k_Base-1;
               reads[n_OverlapReads][7]=1;
               reads2[n_OverlapReads][1]=sinforc[j-1];
               reads2[n_OverlapReads][4]=qinforc[k]+1;
               reads2[n_OverlapReads][5]=qinforc[j-1]+k_Base;
               reads2[n_OverlapReads][6]=sinforc[k]+k_Base-1;
               n_OverlapReads++;
             }
             k=j;
           }
        }

/*      print out the gene names which have been found    */
        if(n_OverlapReads>(MAX_N_BRG-10))
          printf("WARNING: depth too deep and increase seeds!\n");
        if(n_OverlapReads>0)
        {
          int i_index,max_hit,stamt,endmt,sub_len,ii;
          int t_mat,flag_combin,q_min,q_max,s_min,s_max;
          if(q_length<y_len)
          {
            for(i=0;i<(n_OverlapReads-1);i++)
            {
               n_freq=1;
               if((reads[i+1][0]-reads[i][0])==0)
               {
                 k=i+1;
                 n_freq=1;
                 max_hit=reads[i][2];
                 i_index=k-1;
                 while((k<n_OverlapReads)&&((reads[k][0]-reads[k-1][0])==0))
                 {
                   reads[k][7]=0;
                   reads[k-1][7]=0;
                   if(reads[k][2]>max_hit)
                   {
                     max_hit= reads[k][2];
                     i_index=k; 
                   }
                   k++;
                   n_freq++;
                 }
                 if(n_freq>1)
                   reads[i_index][7]=1;
                 i=k-1; 
               }
            }
          }
          else
          {
            memset(mask_que,0,maxq_len*4);
            for(i=0;i<(n_OverlapReads-1);i++)
            {
               n_freq=1;
               if((reads[i+1][0]-reads[i][0])==0)
               {
                 k=i+1;
                 n_freq=1;
                 max_hit=reads[i][2];
                 i_index=k-1;
                 q_min=reads[i][4];
                 s_min=reads[i][1];
                 q_max=reads[i][5];
                 s_max=reads[i][6];
                 t_mat=reads[i][2];
                 while((k<n_OverlapReads)&&((reads[k][0]-reads[k-1][0])==0))
                 {
                   reads[k][7]=0;
                   reads[k-1][7]=0;
                   if(reads[k][2]>max_hit)
                   {
                     max_hit= reads[k][2];
                     i_index=k; 
                   }
                   if(reads[k][4]<q_min)
                     q_min=reads[k][4];
                   if(reads[k][1]<s_min)
                     s_min=reads[k][1];
                   if(reads[k][5]>q_max)
                     q_max=reads[k][5];
                   if(reads[k][6]>s_max)
                     s_max=reads[k][6];
                   t_mat=t_mat+reads[k][2];
                   k++;
                   n_freq++;
                 }
                 flag_combin=0;
                 if(n_freq>1)
                 {
                   if(((q_max-q_min)<50000)&&((s_max-s_min)<50000))
                   {
                     flag_combin=1;
                     reads[i_index][4]=q_min;
                     reads[i_index][5]=q_max;
                     reads[i_index][1]=s_min;
                     reads[i_index][6]=s_max;
                     reads[i_index][2]=t_mat;
                     reads[i_index][7]=1;
                   }
                 }
                 if((n_freq>1)&&(flag_combin==0))
                 {
                   seqs=sub+reads[i_index][0];
                   sub_len=seqs->length;
                   memset(mask_sub,0,sub_len*4);
                   stamt=reads[i_index][4]<100 ? 0 : reads[i_index][4];
                   endmt=(q_length-reads[i_index][5])<100 ? q_length : reads[i_index][5];
//                   printf("end: %d %d\n",stamt,endmt); 
                   for(ii=stamt;ii<endmt;ii++)
                      mask_que[ii]=1;
                   stamt=reads[i_index][1]<100 ? 0 : reads[i_index][1];
                   endmt=(sub_len-reads[i_index][6])<100 ? sub_len : reads[i_index][6]; 
//                   printf("end: %d %d %d\n",stamt,endmt,n_freq); 
                   for(ii=stamt;ii<endmt;ii++)
                      mask_sub[ii]=1;
                   reads[i_index][7]=1;
                   for(jj=(k-n_freq);jj<k;jj++)
                   {
                      if(jj!=i_index)
                      {
                        if((mask_que[reads[jj][4]+100]==1)&&(mask_que[reads[jj][5]-100]==1)&&(mask_sub[reads[jj][1]+100]==1)&&(mask_sub[reads[jj][6]-100]==1))
                          reads[jj][7]=0;
                        else
                        {
                          stamt=reads[jj][4]<100 ? 0 : reads[jj][4];
                          endmt=(q_length-reads[jj][5])<100 ? q_length : reads[jj][5]; 
                          for(ii=stamt;ii<endmt;ii++)
                             mask_que[ii]=1;
                          stamt=reads[jj][1]<100 ? 0 : reads[jj][1];
                          endmt=(sub_len-reads[jj][6])<100 ? sub_len : reads[jj][6]; 
                          for(ii=stamt;ii<endmt;ii++)
                             mask_sub[ii]=1;
                          reads[jj][7]=1;
                        } 
                      }
                   }
                 }
                 i=k-1; 
               }
            }
          }
        }
        if(n_OverlapReads>0) 
        {
          for(i=0;i<n_OverlapReads;i++)
          {
             hit_length[i]=reads[i][2];
             hit_index[i]=i;
          }
          ArraySort2_Int2(n_OverlapReads,hit_length,hit_index);
          for(i=(num_depth+1);i<n_OverlapReads;i++)
             reads[hit_index[i]][7]=0;
        }
        if(n_OverlapReads>0) 
        {
          read1st_flag=0;
//          printf("----------------------------------------------------------------\n");
//          printf("Sequence Name          Index Length Match  RC  Q_start  Q_end  S_start  S_end \n");
          for(i=0;i<n_OverlapReads;i++)
          {
           int cut_query,cut_subjt,mat_query,mat_subjt;
           int mat_len=reads[i][2]*k_Base;
           int mat1_len=reads[i][5]-reads[i][4];
           int mat2_len=reads[i][6]-reads[i][1];
           mat_query=mat_len>(0.1*mat1_len);
/*     needs to think about it ====== mat_subjt      */
           if(sense_flag)
             mat_subjt=1;
           else
             mat_subjt=mat_len>(0.1*mat2_len);
           if((reads[i][7]==1)&&mat_query&&mat_subjt)
           {
            int sub_len,seqq_len,seqs_len,mat_q_offset,mat_s_offset,ic;
            int edge2=0;
            mat_q_offset=0;
            mat_s_offset=0;
            if((!selfcom_flag)||(iseq!=reads[i][0]))
            {
             nline=reads[i][0];
             if(reads[i][3]==0)
             {
               signRC='+';
             }
             else 
               signRC='C';
             seqs=sub+reads[i][0];
//             printf("%-20s   %-6d %-5d %-6d %c   %-6d   %-6d %-6d   %d\n",seqs->name,reads[i][0],ns,reads[i][2],signRC,reads[i][4],reads[i][5],reads[i][1],reads[i][6]);
//             printf("%-20s   %s %-6d %-5d %-6d %c   %-6d   %-6d %-6d   %d\n",seqq->name,seqs->name,reads[i][0],ns,reads[i][2],signRC,reads[i][4],reads[i][5],reads[i][1],reads[i][6]);
             sub_len=seqs->length;
             edge2=edge_len+1000;
             cut_query=(reads[i][4]>edge2)||((q_length-reads[i][5])>edge2);
             cut_subjt=(reads[i][1]>edge2)||((sub_len-reads[i][6])>edge2);
//             if((q_length>10000)&&(((reads[i][5]-reads[i][4])/q_length)<0.1))
             if((q_length>x_len)&&(cut_query))
             {
               memset(quelong,'\0',max_len);
               if(reads[i][4]<=edge_len)
               {
                 mat_q_offset=1;
                 seqq_len=reads[i][5]+edge_len;
               }
               else if((q_length-reads[i][5])<edge_len)
               {
                 mat_q_offset=reads[i][4]-edge_len;
                 seqq_len=q_length-reads[i][4]+edge_len;
               }
               else
               {
                 mat_q_offset=reads[i][4]-edge_len;
                 seqq_len=reads[i][5]-reads[i][4]+2*edge_len;
               }
               if(seqq_len>q_length)
               {
                 seqq_len=q_length;
                 mat_q_offset=0;
               }
               else
               {
                 for(ic=0;ic<seqq_len;ic++)
                 {
                    if(ic>seqq_len-5)
                    {
                      if(seqq->data[mat_q_offset+ic]==' ')
                        quelong[ic]='N';
                      else
                        quelong[ic]=seqq->data[mat_q_offset+ic];
                    }
                    else
                      quelong[ic]=seqq->data[mat_q_offset+ic];
                 }
               }
/*
               printf(">%s\n",seqq->name);
               k=seqq_len/60;
               for(j=0;j<k;j++)
               {
                  for(ic=0;ic<60;ic++)
                     printf("%c",quelong[j*60+ic]);
                  printf("\n");
               }
               for(ic=0;ic<(seqq_len-k*60);ic++)
                  printf("%c",quelong[k*60+ic]);
               printf("\n");
                         */

             }
             else
             {
               seqq_len=seqq->length;
               for(ic=(seqq_len-5);ic<seqq_len;ic++)
               {
                  if(seqq->data[ic]==' ')
                    seqq->data[ic]='N';
               }
             }

             if((sub_len>x_len)&&(cut_subjt))
             {
//               printf("www: %d %d %d %d %d %d\n",sub_len,x_len,cut_subjt,reads[i][1],reads[i][6],edge_len);
               memset(sublong,'\0',max_len);
               if(reads[i][1]<=edge_len)
               {
                 mat_s_offset=1;
                 seqs_len=reads[i][6]+edge_len;
               }
               else if((sub_len-reads[i][6])<edge_len)
               {
                 mat_s_offset=reads[i][1]-edge_len;
                 seqs_len=sub_len-reads[i][1]+edge_len;
               }
               else
               {
                 mat_s_offset=reads[i][1]-edge_len;
                 seqs_len=reads[i][6]-reads[i][1]+2*edge_len;
                 if(seqs_len<(2*edge_len))
                 {
                   mat_s_offset=reads[i][6]-edge_len;
                   seqs_len=reads[i][1]-reads[i][6]+2*edge_len;
                 }
//               printf("www2: %d %d %d %d %d %d\n",sub_len,reads[i][1],reads[i][6],reads[i][3],seqs_len,mat_s_offset);
                 
               }
               if(seqs_len>sub_len)
               {
                 seqs_len=sub_len;
                 mat_s_offset=0;
               }
               else
               {
                 for(ic=0;ic<seqs_len;ic++)
                 {
                    if(ic>seqs_len-5)
                    {
                      if(seqs->data[mat_s_offset+ic]==' ')
                        sublong[ic]='N';
                      else
                        sublong[ic]=seqs->data[mat_s_offset+ic];
                    }
                    else
                      sublong[ic]=seqs->data[mat_s_offset+ic];
                 }
               } 
/*
               printf(">%s\n",seqs->name);
               k=seqs_len/60;
               for(j=0;j<k;j++)
               {
                  for(ic=0;ic<60;ic++)
                     printf("%c",sublong[j*60+ic]);
                  printf("\n");
               }
               for(ic=0;ic<(seqs_len-k*60);ic++)
                  printf("%c",sublong[k*60+ic]);
               printf("\n");
                         */
             }
             else
             {
               seqs_len=seqs->length;
               for(ic=(seqs_len-5);ic<seqs_len;ic++)
               {
                  if(seqs->data[ic]==' ')
                    seqs->data[ic]='N';
               }
             }
             if(debug_flag==1)
               printf("ssaha: %d %d %d %s %s %d %d %d %d %d %d %d %d %d %d %d\n",i,seqq->length,seqs->length,seqq->name,seqs->name,seqq_len,seqs_len,mat_q_offset,mat_s_offset,reads[i][4],reads[i][5],reads[i][1],reads[i][6],reads[i][3],reads[i][2],n_OverlapReads);
//             printf("\n");
             cross_parameters[0]=read1st_flag;
             cross_parameters[1]=info_out;
             cross_parameters[2]=align_flag;
             cross_parameters[3]=mat_q_offset;
             cross_parameters[4]=mat_s_offset;
             cross_parameters[5]=seqq_len;
             cross_parameters[6]=seqs_len;
             cross_parameters[7]=tags_flag;
             cross_parameters[9]=iseq;
             cross_parameters[10]=min_score;
             cross_parameters[11]=align_mem;
             cross_parameters[12]=reads[i][0];
             cross_parameters[13]=identy*100;
             cross_parameters[14]=best_flag;
             if(debug_flag==1)
               printf("before: %d %d %d %d\n",mat_q_offset,mat_s_offset,seqq_len,seqs_len);
             if((mat_q_offset>=0)&&(mat_s_offset>=0))
               cross_match(seqq,seqs,cross_parameters,quelong,sublong,align,&num_align);
             if(debug_flag==1)
               printf("after:\n");
             cross_parameters[8]=1;
/*             printf("Result #       Q_Name             S_Name          Q_Start  Q_End  S_Start  S_End Direction # Bases identity\n");         
             for(i=0;i<num_align;i++)
             {  int match_len = align[i][2]-align[i][1]+1;
                float rate = 100.0 -((align[i][6]+align[i][7]+align[i][8])*100/match_len);
                printf("%d      %20.20s %20.20s %7d %7d %7d %7d     %2.2s    %7d  %5.2f\n",i,seqq->name,seqs->name,align[i][1],align[i][2],align[i][3],align[i][4],"FF",match_len,rate); 
             }   */
             read1st_flag=1;
            }
           }
          }
        }
        if((best_flag)&&(num_align>0))
        {
          int i_align=align[0][13];
          int match_len = align[i_align][2]-align[i_align][1]+1;
          float rate = 100.0 -((align[i_align][6]+align[i_align][7]+align[i_align][8])*100/match_len);
          seqs=sub+align[1][13];

          if(align[i_align][5])
//            tags_flag ? "ALIGNMENT  " : "",
            printf("%s%-5d %s %s %8d %8d %9d %9d   %c %7d %5.2f %d\n",tags_flag ? "ALIGNMENT  " : "",align[i_align][0],seqq->name,seqs->name,align[i_align][1],align[i_align][2],align[i_align][3],align[i_align][4],'C',match_len,rate,seqq->length);
          else 
            printf("%s%-5d %s %s %8d %8d %9d %9d   %c %7d %5.2f %d\n",tags_flag ? "ALIGNMENT  " : "",align[i_align][0],seqq->name,seqs->name,align[i_align][1],align[i_align][2],align[i_align][3],align[i_align][4],'F',match_len,rate,seqq->length);
        }
       }
     }
}


/*   Subroutine to trace read pairs    */
/* ========================================================================== */
void Get_ReadPairs(int rd,int read_pair[Max_N_Pair],int *pair_num)
/* ========================================================================== */
{
     int sign,ptf,ptf1,ptb,ptb1,n_pair;

     n_pair=0;
     if((readpair_f[rd]!=0)&&(readpair_b[rd]!=0))
     {
       sign=readpair_b[rd];
       ptf=abs(readpair_f[rd]);
       while(ptf!=0)
       {
         ptf1=readpair_f[ptf];
         if(ptf>0)
         {
           if(sign>0&&(readpair_b[abs(ptf)]<0))
           {  
             read_pair[n_pair]=ptf;
             n_pair++;
           } 
           else if(sign<0&&(readpair_b[abs(ptf)]>0))
           {
             read_pair[n_pair]=ptf;
             n_pair++;
           }
         }
         ptf=abs(readpair_f[ptf]);
       }
       ptb=abs(readpair_b[rd]);
       while(ptb!=0)
       {
         ptb1=readpair_b[ptb];
         if(ptb>0)
         {
           if(sign>0&&(readpair_f[abs(ptb)]<0))
           {  
             read_pair[n_pair]=ptb;
             n_pair++;
           } 
           else if(sign<0&&(readpair_f[abs(ptb)]>0))
           {
             read_pair[n_pair]=ptb;
             n_pair++;
           }
         }
         ptb=abs(readpair_b[ptb]);
       }
     }
     else if((readpair_f[rd]==0)&&(readpair_b[rd]!=0))
     {
       sign=readpair_b[rd];
       ptb=abs(readpair_b[rd]);
       while(ptb!=0)
       {
         ptb1=readpair_b[ptb];
         if(ptb>0)
         {
           if(sign>0&&(readpair_f[abs(ptb)]<0))
           {  
             read_pair[n_pair]=ptb;
             n_pair++;
           } 
           else if(sign<0&&(readpair_f[abs(ptb)]>0))
           {
             read_pair[n_pair]=ptb;
             n_pair++;
           }
         }
         ptb=abs(readpair_b[ptb]);
       }
     }
     else if((readpair_f[rd]!=0)&&(readpair_b[rd]==0))
     {
       sign=readpair_f[rd];
       ptf=abs(readpair_f[rd]);
       while(ptf!=0)
       {
         ptf1=readpair_f[ptf];
         if(ptf>0)
         {
           if(sign>0&&(readpair_b[abs(ptf)]<0))
           {  
             read_pair[n_pair]=ptf;
             n_pair++;
           } 
           else if(sign<0&&(readpair_b[abs(ptf)]>0))
           {
             read_pair[n_pair]=ptf;
             n_pair++;
           }
         }
         ptf=abs(readpair_f[ptf]);
       }
     }
     *pair_num=n_pair;
}

/* ============== */
void memory_setup() 
/* ============= */
{
   if((fdata = (SIO *)calloc(nfm,sizeof(SIO))) == NULL)
   {
     printf("ERROR Frequence array: calloc - fdata\n");
     exit(1);
   }
   if((fdataRC = (SIO *)calloc(nfm,sizeof(SIO))) == NULL)
   {
     printf("ERROR Frequence array: calloc - fdataRC\n");
     exit(1);
   }
   if((qinfo = (int *)calloc(nfm,sizeof(int))) == NULL)
   {
     printf("ERROR Frequence array: calloc - qinfo\n");
     exit(1);
   }
   if((qinforc = (int *)calloc(nfm,sizeof(int))) == NULL)
   {
     printf("ERROR Frequence array: calloc - qinforc\n");
     exit(1);
   }
   if((sinfo = (int *)calloc(nfm,sizeof(int))) == NULL)
   {
     printf("ERROR Frequence array: calloc - sinfo\n");
     exit(1);
   }
   if((sinforc = (int *)calloc(nfm,sizeof(int))) == NULL)
   {
     printf("ERROR Frequence array: calloc - sinforc\n");
     exit(1);
   }
}

/* =============================== */
void information()
/* =============================== */
{
   int i,j;

   printf("DESCRIPTION\n");
   printf("\n");
   printf(" ssaha is a software tool for very fast matching and alignment of DNA sequences.\n"); 
   printf(" It stands for Sequence Search and Alignment by Hashing Algorithm. It achieves\n");
   printf(" its fast search speed by converting sequence information into a `hash table'\n");
   printf(" data structure, which can then be searched very rapidly for matches.\n");
   printf("\n");
   printf(" ssaha2 is a package combined of ssaha with phrap/cross_match developed by\n");
   printf(" Phil Green at the University of Washington. Matching seeds of a few kmer words\n");
   printf(" are found by the ssaha algorithm. Both query and subject sequences are cut off\n");
   printf(" according to the locations of the matching seeds and then passed to cross_match\n");
   printf(" for full alignment.\n");
   printf("\n");
   printf("OPTIONS\n");
   printf("\n");
   printf(" -kmer:     length of kmer words;\n");
   printf("            default kmer=12\n");
   printf(" -seeds:    number of exact kmer words match in the subject, as matching seeds;\n");
   printf("            default seeds=10\n");
   printf(" -align:    '1' - show full alignment; '0' - no alignment;\n");
   printf("            default '1'\n");
   printf(" -sense:    '1' - search with higher sensitivity; '0' - normal;\n");
   printf("            default '0'\n");
   printf(" -tags:     '1' - show a tag of 'ALIGNMENT'; '0' - no tag;\n");
   printf("            default '0'\n");
   printf(" -depth:    number of reported hits with best alignment;\n");
   printf("            default depth=50\n");
   printf(" -score:    minimum score of smith-waterman, below which hits are not reported;\n");
   printf("            default score=30\n");
   printf(" -cut:      number of word occurrence in the dataset, above which the words are\n");
   printf("            ignored; default cut=200\n");
   printf(" -memory:   memory assigned in MBs for cross_match alignment;\n");
   printf("            default memory=2000\n");
   printf(" -array:    memory assigned in MBs for storing frequence arrays;\n");
   printf("            default memory=4\n");
   printf(" -edge:     extension of both ends on the subject sequence to pick up small hits;\n");
   printf("            default edge=200\n");
   printf(" -best:     report the best alignment from the hit list;\n");
   printf("            default '0'\n");
   printf(" -start:    start read from the query file - search starts from this read;\n");
   printf("            default start=0\n");
   printf(" -end:      end read from the query file - search ends at this read;\n");
   printf("            default start= Total number of the reads in the query file;\n");
   printf("\n");
   printf("COOKBOOK\n");
   printf("\n");
   printf(" To get the best results from the program, you may tune around the parameters.\n");
   printf(" To find the best hit in the database, use higher values of seeds, score and\n");
   printf(" lower depth value. Small values in score and seeds, plus -sense 1 option will\n");
   printf(" increase sensitivity, but more CPU time. The parameter of depth helps you to\n");
   printf(" control the output results. Kmer is also very important and it should be set\n");
   printf(" to 13 or 14 if the database is more than 1 GB. In the following, we recommand\n");
   printf(" a set of parameters for some applications:\n"); 
   printf("\n");
   printf(" BACends placement  - find the best hit in the database:        \n");
   printf("    -seeds 14 -kmer 13 -align 0 -tags 1 -depth 5 -score 200;\n");
   printf("\n");
   printf(" EST/cDNA alignment - produce splice on the subject sequence: \n");
   printf("    -seeds 4 -kmer 13 -align 0 -tags 1 -depth 5 -score 20 -edge 20000;\n");
   printf("\n");
   printf(" Primer/gene Marks alignment - find the matches of short motifs to the database:\n");
   printf("    -seeds 1 -kmer 13 -tags 1 -score 12 -skip 1 -sense 1 -cut 50000;\n");
   printf("\n");
   printf(" Search with higher sensitivity:\n");
   printf("    -seeds 2 -kmer 13 -tags 1 -score 20 -sense 1 -cut 50000;\n");
   printf("\n");
   printf(" Both query and subject are large (q: 100Kb < query < 1MB; s: no limit):\n");
   printf("    -seeds 50 -kmer 13 -tags 1 -score 2000 -array 40 -memory 10000;\n");
   printf("\n");
   printf("WARNING\n");
   printf("\n");
   printf(" For searches with both long query and long subject sequences, you need to\n");
   printf(" assign big memory for cross_match. Otherwise, the code might crash!\n"); 
   printf("\n");
}

#define SWAP(a,b) temp=(a);(a)=b;(b)=temp;

/*   Subroutine to sort an array arr[0,...,n-1] into ascending order while
     making the corresponding reaarangement of the array brr[0,...,n-1]
     by the use of Quicksort (Sedgwick, R. 1978, Communications o fthe ACM,
     vol. 21, pp. 847-857) also see Numerical Recipes in C                  */  

/* =============================== */
void ArraySort_Long(int n, int64_t *arr)
/* =============================== */
{
     int i,ir=n-1,j,k,m=0,jstack=0,NSTACK=50,istack[NSTACK];
     int64_t a,temp,MIN=7;

     for(;;)
     {
/*      Insertion sort when subarray is small enough    */
        if(ir-m<MIN)
        {
          for(j=m+1;j<=ir;j++)
          {
             a=arr[j];
             for(i=j-1;i>=m;i--)
             {
                if(arr[i]<=a) break;
                arr[i+1]=arr[i];
             }
             arr[i+1]=a;
          }
          if(!jstack) return;
          ir=istack[jstack--];
          m=istack[jstack--];
        }
        else
        {
          k=(m+ir)>>1;
          SWAP(arr[k],arr[m+1]);

          if(arr[m]>arr[ir])
          {
            SWAP(arr[m],arr[ir]);
          }

          if(arr[m+1]>arr[ir])
          {
            SWAP(arr[m+1],arr[ir]);
          }

          if(arr[m]>arr[m+1])
          {
            SWAP(arr[m],arr[m+1]);
          }

          i=m+1;
          j=ir;
          a=arr[m+1];
          for(;;)
          {
             do i++; while (arr[i]<a);
             do j--; while (arr[j]>a);
             if(j<i) break;
             SWAP(arr[i],arr[j]);
          }
          arr[m+1]=arr[j];
          arr[j]=a;
          jstack+=2;

/*        Push pointers to larger subarray on stack      */
/*        process smaller subarray immediately           */
          if(jstack>NSTACK)
          {
             printf("Stack error: NSTACK too small\n");
             exit(0);
          }
          if(ir-i+1>=j-m)
          {
            istack[jstack]=ir;
            istack[jstack-1]=i;
            ir=j-1;
          }
          else
          {
            istack[jstack]=j-1;
            istack[jstack-1]=m;
            m=i;
          }
        }
     }
}


/* =============================== */
void ArraySort_Int(int n, int *arr)
/* =============================== */
{
     int i,ir=n-1,j,k,m=0,jstack=0,NSTACK=50,istack[NSTACK];
     int a,temp,MIN=7;

     for(;;)
     {
/*      Insertion sort when subarray is small enough    */
        if(ir-m<MIN)
        {
          for(j=m+1;j<=ir;j++)
          {
             a=arr[j];
             for(i=j-1;i>=m;i--)
             {
                if(arr[i]<=a) break;
                arr[i+1]=arr[i];
             }
             arr[i+1]=a;
          }
          if(!jstack) return;
          ir=istack[jstack--];
          m=istack[jstack--];
        }
        else
        {
          k=(m+ir)>>1;
          SWAP(arr[k],arr[m+1]);

          if(arr[m]>arr[ir])
          {
            SWAP(arr[m],arr[ir]);
          }

          if(arr[m+1]>arr[ir])
          {
            SWAP(arr[m+1],arr[ir]);
          }

          if(arr[m]>arr[m+1])
          {
            SWAP(arr[m],arr[m+1]);
          }

          i=m+1;
          j=ir;
          a=arr[m+1];
          for(;;)
          {
             do i++; while (arr[i]<a);
             do j--; while (arr[j]>a);
             if(j<i) break;
             SWAP(arr[i],arr[j]);
          }
          arr[m+1]=arr[j];
          arr[j]=a;
          jstack+=2;

/*        Push pointers to larger subarray on stack      */
/*        process smaller subarray immediately           */
          if(jstack>NSTACK)
          {
             printf("Stack error: NSTACK too small\n");
             exit(0);
          }
          if(ir-i+1>=j-m)
          {
            istack[jstack]=ir;
            istack[jstack-1]=i;
            ir=j-1;
          }
          else
          {
            istack[jstack]=j-1;
            istack[jstack-1]=m;
            m=i;
          }
        }
     }
}


/* =============================== */
void ArraySort_Mix(int n, int64_t *arr, int *brr)
/* =============================== */
{
     int i,ir=n-1,j,k,m=0,jstack=0,b,NSTACK=50,istack[NSTACK];
     int64_t a,temp,MIN=7;

     for(;;)
     {
/*      Insertion sort when subarray is small enough    */
        if(ir-m<MIN)
        {
          for(j=m+1;j<=ir;j++)
          {
             a=arr[j];
             b=brr[j];
             for(i=j-1;i>=m;i--)
             {
                if(arr[i]<=a) break;
                arr[i+1]=arr[i];
                brr[i+1]=brr[i];
             }
             arr[i+1]=a;
             brr[i+1]=b;
          }
          if(!jstack) return;
          ir=istack[jstack--];
          m=istack[jstack--];
        }
        else
        {
          k=(m+ir)>>1;
          SWAP(arr[k],arr[m+1]);
          SWAP(brr[k],brr[m+1]);

          if(arr[m]>arr[ir])
          {
            SWAP(arr[m],arr[ir]);
            SWAP(brr[m],brr[ir]);
          }

          if(arr[m+1]>arr[ir])
          {
            SWAP(arr[m+1],arr[ir]);
            SWAP(brr[m+1],brr[ir]);
          }

          if(arr[m]>arr[m+1])
          {
            SWAP(arr[m],arr[m+1]);
            SWAP(brr[m],brr[m+1]);
          }

          i=m+1;
          j=ir;
          a=arr[m+1];
          b=brr[m+1];
          for(;;)
          {
             do i++; while (arr[i]<a);
             do j--; while (arr[j]>a);
             if(j<i) break;
             SWAP(arr[i],arr[j]);
             SWAP(brr[i],brr[j]);
          }
          arr[m+1]=arr[j];
          arr[j]=a;
          brr[m+1]=brr[j];
          brr[j]=b;
          jstack+=2;

/*        Push pointers to larger subarray on stack      */
/*        process smaller subarray immediately           */
          if(jstack>NSTACK)
          {
             printf("Stack error: NSTACK too small\n");
             exit(0);
          }
          if(ir-i+1>=j-m)
          {
            istack[jstack]=ir;
            istack[jstack-1]=i;
            ir=j-1;
          }
          else
          {
            istack[jstack]=j-1;
            istack[jstack-1]=m;
            m=i;
          }
        }
     }
}

/* =============================== */
void ArraySort_Int2(int n, int *arr, int *brr)
/* =============================== */
{
     int i,ir=n-1,j,k,m=0,jstack=0,b,NSTACK=50,istack[NSTACK];
     int a,temp,MIN=7;

     for(;;)
     {
/*      Insertion sort when subarray is small enough    */
        if(ir-m<MIN)
        {
          for(j=m+1;j<=ir;j++)
          {
             a=arr[j];
             b=brr[j];
             for(i=j-1;i>=m;i--)
             {
                if(arr[i]<=a) break;
                arr[i+1]=arr[i];
                brr[i+1]=brr[i];
             }
             arr[i+1]=a;
             brr[i+1]=b;
          }
          if(!jstack) return;
          ir=istack[jstack--];
          m=istack[jstack--];
        }
        else
        {
          k=(m+ir)>>1;
          SWAP(arr[k],arr[m+1]);
          SWAP(brr[k],brr[m+1]);

          if(arr[m]>arr[ir])
          {
            SWAP(arr[m],arr[ir]);
            SWAP(brr[m],brr[ir]);
          }

          if(arr[m+1]>arr[ir])
          {
            SWAP(arr[m+1],arr[ir]);
            SWAP(brr[m+1],brr[ir]);
          }

          if(arr[m]>arr[m+1])
          {
            SWAP(arr[m],arr[m+1]);
            SWAP(brr[m],brr[m+1]);
          }

          i=m+1;
          j=ir;
          a=arr[m+1];
          b=brr[m+1];
          for(;;)
          {
             do i++; while (arr[i]<a);
             do j--; while (arr[j]>a);
             if(j<i) break;
             SWAP(arr[i],arr[j]);
             SWAP(brr[i],brr[j]);
          }
          arr[m+1]=arr[j];
          arr[j]=a;
          brr[m+1]=brr[j];
          brr[j]=b;
          jstack+=2;

/*        Push pointers to larger subarray on stack      */
/*        process smaller subarray immediately           */
          if(jstack>NSTACK)
          {
             printf("Stack error: NSTACK too small\n");
             exit(0);
          }
          if(ir-i+1>=j-m)
          {
            istack[jstack]=ir;
            istack[jstack-1]=i;
            ir=j-1;
          }
          else
          {
            istack[jstack]=j-1;
            istack[jstack-1]=m;
            m=i;
          }
        }
     }
}

/*   function to sort an array into a decreasing order:  a>b>c>....    */  
/* =============================== */
void ArraySort2_Int2(int n, int *arr, int *brr)
/* =============================== */
{
     int i,ir=n-1,j,k,m=0,jstack=0,b,NSTACK=50,istack[NSTACK];
     int a,temp,MIN=7;

     for(;;)
     {
/*      Insertion sort when subarray is small enough    */
        if(ir-m<MIN)
        {
          for(j=m+1;j<=ir;j++)
          {
             a=arr[j];
             b=brr[j];
             for(i=j-1;i>=m;i--)
             {
                if(arr[i]>=a) break;
                arr[i+1]=arr[i];
                brr[i+1]=brr[i];
             }
             arr[i+1]=a;
             brr[i+1]=b;
          }
          if(!jstack) return;
          ir=istack[jstack--];
          m=istack[jstack--];
        }
        else
        {
          k=(m+ir)>>1;
          SWAP(arr[k],arr[m+1]);
          SWAP(brr[k],brr[m+1]);

          if(arr[m]<arr[ir])
          {
            SWAP(arr[m],arr[ir]);
            SWAP(brr[m],brr[ir]);
          }

          if(arr[m+1]<arr[ir])
          {
            SWAP(arr[m+1],arr[ir]);
            SWAP(brr[m+1],brr[ir]);
          }

          if(arr[m]<arr[m+1])
          {
            SWAP(arr[m],arr[m+1]);
            SWAP(brr[m],brr[m+1]);
          }

          i=m+1;
          j=ir;
          a=arr[m+1];
          b=brr[m+1];
          for(;;)
          {
             do i++; while (arr[i]>a);
             do j--; while (arr[j]<a);
             if(j<i) break;
             SWAP(arr[i],arr[j]);
             SWAP(brr[i],brr[j]);
          }
          arr[m+1]=arr[j];
          arr[j]=a;
          brr[m+1]=brr[j];
          brr[j]=b;
          jstack+=2;

/*        Push pointers to larger subarray on stack      */
/*        process smaller subarray immediately           */
          if(jstack>NSTACK)
          {
             printf("Stack error: NSTACK too small\n");
             exit(0);
          }
          if(ir-i+1>=j-m)
          {
            istack[jstack]=ir;
            istack[jstack-1]=i;
            ir=j-1;
          }
          else
          {
            istack[jstack]=j-1;
            istack[jstack-1]=m;
            m=i;
          }
        }
     }
}

/* =============================== */
void ArraySort_Mix3(int n, int64_t *arr, int *brr, int *crr)
/* =============================== */
{
     int i,ir=n-1,j,k,m=0,jstack=0,b,c,NSTACK=50,istack[NSTACK];
     int64_t a,temp,MIN=7;

     for(;;)
     {
/*      Insertion sort when subarray is small enough    */
        if(ir-m<MIN)
        {
          for(j=m+1;j<=ir;j++)
          {
             a=arr[j];
             b=brr[j];
             c=crr[j];
             for(i=j-1;i>=m;i--)
             {
                if(arr[i]<=a) break;
                arr[i+1]=arr[i];
                brr[i+1]=brr[i];
                crr[i+1]=crr[i];
             }
             arr[i+1]=a;
             brr[i+1]=b;
             crr[i+1]=c;
          }
          if(!jstack) return;
          ir=istack[jstack--];
          m=istack[jstack--];
        }
        else
        {
          k=(m+ir)>>1;
          SWAP(arr[k],arr[m+1]);
          SWAP(brr[k],brr[m+1]);
          SWAP(crr[k],crr[m+1]);

          if(arr[m]>arr[ir])
          {
            SWAP(arr[m],arr[ir]);
            SWAP(brr[m],brr[ir]);
            SWAP(crr[m],crr[ir]);
          }

          if(arr[m+1]>arr[ir])
          {
            SWAP(arr[m+1],arr[ir]);
            SWAP(brr[m+1],brr[ir]);
            SWAP(crr[m+1],crr[ir]);
          }

          if(arr[m]>arr[m+1])
          {
            SWAP(arr[m],arr[m+1]);
            SWAP(brr[m],brr[m+1]);
            SWAP(crr[m],crr[m+1]);
          }

          i=m+1;
          j=ir;
          a=arr[m+1];
          b=brr[m+1];
          c=crr[m+1];
          for(;;)
          {
             do i++; while (arr[i]<a);
             do j--; while (arr[j]>a);
             if(j<i) break;
             SWAP(arr[i],arr[j]);
             SWAP(brr[i],brr[j]);
             SWAP(crr[i],crr[j]);
          }
          arr[m+1]=arr[j];
          arr[j]=a;
          brr[m+1]=brr[j];
          brr[j]=b;
          crr[m+1]=crr[j];
          crr[j]=c;
          jstack+=2;

/*        Push pointers to larger subarray on stack      */
/*        process smaller subarray immediately           */
          if(jstack>NSTACK)
          {
             printf("Stack error: NSTACK too small\n");
             exit(0);
          }
          if(ir-i+1>=j-m)
          {
            istack[jstack]=ir;
            istack[jstack-1]=i;
            ir=j-1;
          }
          else
          {
            istack[jstack]=j-1;
            istack[jstack-1]=m;
            m=i;
          }
        }
     }
}


/*   to swap the string arrays           */
/* ============================================= */
void s_swap(char **Pair_Name, int i, int j)
/* ============================================= */
{
     char temp[Max_N_NameBase];

     strcpy(temp,Pair_Name[i]);
     strcpy(Pair_Name[i],Pair_Name[j]);
     strcpy(Pair_Name[j],temp);
}


/*   to sort the string array in order          */
/* ============================================= */
void ArraySort_String(int n, char **Pair_Name, int *brr)
/* ============================================= */
{
     int i,ir=n-1,j,k,m=0,jstack=0,b,NSTACK=50,istack[NSTACK];
     int temp,MIN=7;
     char p[Max_N_NameBase];

     for(;;)
     {
/*      Insertion sort when subarray is small enough    */
        if(ir-m<MIN)
        {
          for(j=m+1;j<=ir;j++)
          {
             strcpy(p,Pair_Name[j]);
             b=brr[j];
             for(i=j-1;i>=m;i--)
             {
                if(strcmp(Pair_Name[i],p)<=0) break;
                strcpy(Pair_Name[i+1],Pair_Name[i]);
                brr[i+1]=brr[i];
             }
             strcpy(Pair_Name[i+1],p);
             brr[i+1]=b;
          }
          if(!jstack) return;
          ir=istack[jstack--];
          m=istack[jstack--];
        }
        else
        {
          k=(m+ir)>>1;
          s_swap(Pair_Name,k,m+1);
          SWAP(brr[k],brr[m+1]);

          if(strcmp(Pair_Name[m],Pair_Name[ir])>0)
          {
            s_swap(Pair_Name,m,ir);
            SWAP(brr[m],brr[ir]);
          }

          if(strcmp(Pair_Name[m+1],Pair_Name[ir])>0)
          {
            s_swap(Pair_Name,m+1,ir);
            SWAP(brr[m+1],brr[ir]);
          }

          if(strcmp(Pair_Name[m],Pair_Name[m+1])>0)
          {
            s_swap(Pair_Name,m,m+1);
            SWAP(brr[m],brr[m+1]);
          }

          i=m+1;
          j=ir;
          strcpy(p,Pair_Name[m+1]);
          b=brr[m+1];
          for(;;)
          {
             do i++; while (strcmp(Pair_Name[i],p)<0);
             do j--; while (strcmp(Pair_Name[j],p)>0);
             if(j<i) break;
             s_swap(Pair_Name,i,j);
             SWAP(brr[i],brr[j]);
          }
          strcpy(Pair_Name[m+1],Pair_Name[j]);
          strcpy(Pair_Name[j],p);
          brr[m+1]=brr[j];
          brr[j]=b;
          jstack+=2;

/*        Push pointers to larger subarray on stack      */
/*        process smaller subarray immediately           */
          if(jstack>NSTACK)
          {
             printf("Stack error: NSTACK too small\n");
             exit(0);
          }
          if(ir-i+1>=j-m)
          {
            istack[jstack]=ir;
            istack[jstack-1]=i;
            ir=j-1;
          }
          else
          {
            istack[jstack]=j-1;
            istack[jstack-1]=m;
            m=i;
          }
        }
     }
}


/* creat an int matrix with subscript ange m[nrl...nrh][ncl...nch]  */
int     **imatrix(int64_t nrl,int64_t nrh,int64_t ncl,int64_t nch)
{
        int64_t i, nrow=nrh-nrl+1,ncol=nch-ncl+1;
        int  **m;

        /* allocate pointers to rows        */
        if((m=(int **)calloc(nrow,sizeof(int*)))==NULL)
        {
           printf("error imatrix: calloc error No. 1 \n");
           return(NULL);
        }
        m+=0;
        m-=nrl;

        /* allocate rows and set pointers to them        */
        if((m[nrl]=(int *)calloc(nrow*ncol,sizeof(int)))==NULL)
        {
           printf("error imatrix: calloc error No. 2 \n");
           return(NULL);
        }
        m[nrl]+=0;
        m[nrl]-=ncl;

        for(i=nrl+1;i<=nrh;i++)
           m[i]=m[i-1]+ncol;
        /* return pointer to array of pointers to rows   */
        return m;
}
         
/* creat a char matrix with subscript ange c[nrl...nrh][ncl...nch]  */
char     **cmatrix(long nrl,long nrh,long ncl,long nch)
{
        long i, nrow=nrh-nrl+1,ncol=nch-ncl+1;
        char  **m;

        /* allocate pointers to rows        */
        if((m=(char **)calloc(nrow,sizeof(char*)))==NULL)
        {
           printf("error cmatrix: calloc error No. 1 \n");
           return(NULL);
        }
        m+=0;
        m-=nrl;

        /* allocate rows and set pointers to them        */
        if((m[nrl]=(char *)calloc(nrow*ncol,sizeof(char)))==NULL)
        {
           printf("error cmatrix: calloc error No. 2 \n");
           return(NULL);
        }
        m[nrl]+=0;
        m[nrl]-=nrl;

        for(i=nrl+1;i<=nrh;i++)
           m[i]=m[i-1]+ncol;
        /* return pointer to array of pointers to rows   */
        return m;
}



