/*****************************************************************************
#   Copyright (C) 1993-1999 by Phil Green.                          
#   All rights reserved.                           
#                                                                           
#   This software is part of a beta-test version of the swat/cross_match/phrap 
#   package.  It should not be redistributed or
#   used for any commercial purpose, including commercially funded
#   sequencing, without written permission from the author and the
#   University of Washington.
#   
#   This software is provided ``AS IS'' and any express or implied
#   warranties, including, but not limited to, the implied warranties of
#   merchantability and fitness for a particular purpose, are disclaimed.
#   In no event shall the authors or the University of Washington be
#   liable for any direct, indirect, incidental, special, exemplary, or
#   consequential damages (including, but not limited to, procurement of
#   substitute goods or services; loss of use, data, or profits; or
#   business interruption) however caused and on any theory of liability,
#   whether in contract, strict liability, or tort (including negligence
#   or otherwise) arising in any way out of the use of this software, even
#   if advised of the possibility of such damage.
#                                                                         
#*****************************************************************************/

#include <math.h>
#include <values.h>
#include <stdio.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <dirent.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>

extern int cross_out;
extern int alloc_flag;
extern int newpit_flag;
extern int flag_array[20];

char *plural(int_var)
     int int_var;
{
  return int_var == 1 ? "" : "s";
}

/* N.B. following needs to be tested for non-unix systems */
fatalError(message)
     char *message;
{
  fprintf(stderr, "\nFATAL 888 ERROR: %s\n", message);
  exit(1);
}

notify(message)
     char *message;
{
  fprintf(stderr, message);
  if (!strcmp(message, " Done\n")) check_alloc_list();
  fflush(stderr);
  fflush(stdout);
}

FILE *fopenRead(fileName)
  char *fileName;
{
  FILE *fp;
  FILE *fopen();

  if(!(fp = fopen(fileName,"r"))) {
    fprintf(stderr, "\n ERROR: FILE1 %s NOT FOUND \n",fileName);
    exit(1);
  }
  else return fp;
}

FILE *fopenWrite(fileName)
  char *fileName;
{
  FILE *fp;
  FILE *fopen();

  if(!(fp = fopen(fileName,"w"))) {
    fprintf(stderr, "\n ERROR: FILE2 %s NOT AVAILABLE FOR WRITING \n",fileName);
    exit(1);
  }
  else return fp;
}

FILE *fopenAppend(fileName)
  char *fileName;
{
  FILE *fp;
  FILE *fopen();

  if(!(fp = fopen(fileName,"a"))) {
    fprintf(stderr, "\n ERROR: FILE3 %s NOT AVAILABLE FOR WRITING \n",fileName);
    exit(1);
  }
  else return fp;
}

/* N.B. The following is modified from Kernighan & Ritchie, The C Programming
Language, 1st edition */

typedef unsigned int ALLOC;
typedef short SHORT;

typedef double ALIGN;

union header {
	struct {
		union header *ptr;
		ALLOC size;
	} s;
	ALIGN  x;
};

typedef union header HEADER;

static HEADER base;
static HEADER *shead;
static HEADER *allocp = 0;	
static int64_t total_allocated;
static int head_flag=0;
static int NUM_UNIT=0;
int64_t sizemem=1L*200000000;

char *ssaha_free(nbytes)
{
     HEADER *p;
     HEADER *q;
     free(shead);
     return 0;
}

char *our_alloc(nbytes)
        ALLOC nbytes;
{
	HEADER *morecore();
	HEADER *p;
	HEADER *q;
	ALLOC nunits;
 
	nunits = 1 +(nbytes + sizeof(HEADER) - 1)/ sizeof(HEADER);
        if(alloc_flag==0)
        {
//          *allocp=0;
          NUM_UNIT=0;
          total_allocated=0;
          alloc_flag=1;
        }
        if(head_flag==0)
        {
          sizemem=(1000000L*flag_array[7]);
          sizemem=(1L*sizemem)/sizeof(HEADER);
//          printf("mem: %d %d\n",flag_array[7],sizemem*sizeof(HEADER));
          shead=(HEADER *)malloc(sizemem * sizeof(HEADER));
          head_flag=1;
        }
	if ((!(q = allocp))||(newpit_flag==0)) {
		base.s.ptr = allocp = q = &base;
		base.s.size = 0;
                memset(shead,0,flag_array[18] * sizeof(HEADER));
                newpit_flag=1;
	}

        morecore(nunits);
        p=allocp;
        return ((char *)(p + 1));;
}

check_alloc_list()
{
	int64_t num_blocks, num_ends, num_units;
	HEADER *p, *q;

	p = allocp;
	num_blocks = num_ends = num_units = 0;
        printf("check: %d %d\n",allocp,p);
	do {
	  q = p->s.ptr;
//        printf("check: %d %d %d %d %d %d %d\n",num_blocks,allocp,q,p,num_units,num_ends,p->s.size);
	  if (p->s.size < 0)
	    fatalError(" our_alloc linked list corruption (check_alloc_list1)");
	  num_units += p->s.size;
	  if (q > p && p + p->s.size > q)
	    fatalError(" our_alloc linked list corruption (check_alloc_list2)");
	  if (q < p) num_ends++;
	  p->s.size=0;
	  p->s.ptr=allocp;
	  p = q;
	  num_blocks++;
	} while (p != allocp);
//	if (num_ends != 1)
//	    fatalError(" our_alloc linked list corruption (check_alloc_list3)");
        if(cross_out==1)
	fprintf(stderr, "Total space allocated: %.3f Mbytes; currently free: %.3f Mbytes in %d blocks\n",
		total_allocated / 1000000.0, num_units * sizeof(HEADER) / 1000000.0, num_blocks);
	printf("Total space allocated: %.3f Mbytes; currently free: %.3f Mbytes in %d blocks\n",
		total_allocated / 1000000.0, num_units * sizeof(HEADER) / 1000000.0, num_blocks);
}

zero_alloc_list()
{
	int64_t num_blocks;
	HEADER *p, *q;

	p = allocp;
	num_blocks = 0;
	do {
	  q = p->s.ptr;
        printf("zero: %d %d %d\n",num_blocks,allocp,q);
	  if (p->s.size < 0)
	    fatalError(" our_alloc linked list corruption (check_alloc_list1)");
	  if (q > p && p + p->s.size > q)
	    fatalError(" our_alloc linked list corruption (check_alloc_list2)");
	  p->s.size=0;
	  p->s.ptr=allocp;
	  p = q;
	  num_blocks++;
	} while (p != allocp);
}

find_frags()
{
	SHORT num_frags;
	HEADER *p;

	for (p = allocp->s.ptr, num_frags = 1; p != allocp;
            p = p->s.ptr, num_frags++);
        if(cross_out==1)
        fprintf(stderr, "\n number of allocation fragments = %d", num_frags);
}

#define  NALLOC  63000

HEADER *morecore(nu)
        ALLOC nu;
{
//	char *malloc();
	char *cp,*cp2;
	HEADER *up,*p,*q;
	ALLOC rnu;
        SHORT our_free();
   
//	rnu = NALLOC * ((nu + NALLOC - 1) / NALLOC);
        rnu=nu+10;
//        cp = malloc(rnu * sizeof(HEADER));
	total_allocated += rnu * sizeof(HEADER);
        if(cross_out==1)
          fprintf(stderr, "%.3f Mbytes allocated -- total %.3f Mbytes\n", 
		rnu * sizeof(HEADER) / 1000000.0, total_allocated / 1000000.0);
	up = shead+NUM_UNIT;
        sizemem=(1000000L*flag_array[7]);
        if(NUM_UNIT>(sizemem-1000L))
        {
          flag_array[0]=1;
          return allocp;
        }
//      up = (HEADER *)cp;
	up->s.size = rnu; 
        p=up;
        q=allocp;
        p->s.ptr = q->s.ptr;
        q->s.ptr = p;
        allocp = up;  
        NUM_UNIT=NUM_UNIT+rnu;
        flag_array[18]=NUM_UNIT;
	return allocp;
}

SHORT our_free(ap)
     char *ap;
{
	HEADER *p;
	HEADER *q;
	
	if (!ap) return 0; 

	p = (HEADER *)ap - 1;
        q = allocp;
	allocp = p;
        return 0;
	for (q = allocp; !(p > q && p < q->s.ptr); q = q->s.ptr)
        {
		if (q>= q->s.ptr && (p > q || p < q->s.ptr)) 
                {
                   break;
                }
        }

	if ( p + p->s.size == q->s.ptr)
        {
		p->s.size += q->s.ptr->s.size;
		p->s.ptr = q->s.ptr->s.ptr;
	} 
	else if (p + p->s.size > q->s.ptr && p <= q->s.ptr) 
        {
	  fatalError(" our_alloc linked list corruption (our_free1)");
	}
	else p->s.ptr = q->s.ptr;
	if (q + q->s.size == p){
		q->s.size += p->s.size;
		q->s.ptr = p->s.ptr;
	} 
	else if (q + q->s.size > p && p >= q) 
        {
        printf("free: %d %d %d %d\n",q,p,q->s.size,q + q->s.size);
	  fatalError(" our_alloc linked list corruption (our_free2)");
	}
	else q->s.ptr = p;
	allocp = &base;
        return 0;
}
 
